"Python3 BEAGLE rule-evaluation tester."

##  BERRIES :
##
##  Bionically Evolved Regression Rule In Executable Software.
##  Copyright (c), 2016, R.S. Forsyth.
##  ancestral progenitor : Sept, 1980.
##  first version : 17/12/2015
##  last revision : 13/09/2016


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  copy
import  math
import  os, sys
import  random, time

##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  parashow
from glib import  showobj
from glib import  openfile
from glib import  request
from glib import  pybyebye
from gridlib import  cols2use, matread
from gridlib import  matdump, vecline
from gridlib import  var2col
from gridlib import  bastatz, midvalue
from gridlib import  vecdist
from gridlib import  copygrid, distest
from gridlib import  matshow, precrec
from gridlib import  sckappa2
from gridlib import  hedgesg
from gridlib import  corsums, corrval
from gridlib import  ranx
from rulelib import  *
from root import  targets
from tree import  dumpline


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ##  ("atomize",1,0,1),
    ("randseed",0,0,999999999),  ##  random seed, <2 means haphazard seed
    ("jobname","runster"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ("testdat",''),  ##  file for input of data
    ##  ("cols2use",''),  ##  variables to be used
    ("rulemode", "tabular"),  ##  default is tabular
    ("targval",''),  ##  target variable
    ("targcol",0),  ##  target column number [2b reset]
    ("idfield",''),  ##  varname for o/p label
    ("skipvars",''),  ##  variables to be excluded
    ("listfile",''),  ##  main output file
    ("dumpfile",''),  ##  output file for subsampling results (?)
    ("progfile",''),  ##  file containing Python code
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"berries.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "Bionically Evolved Regression Rule In Executable Software v3.0"
##  Gold = 5.0**0.5*0.5+0.5
Mincases = 2
Dattail = "_dat2"
Dumptail = "_pout"
Listtail = "_prop"  ##  program o/p
Progtail = "_prog.py"
Maxcats = 64
Dustbin = "??!!"
Goodsign = '+'
Poorsign = '-'
Star = '*'
Longname = 26  ##  length after which category names will be truncated
Longlab = 22  ##  length after which label fields will be truncated
Outpars = ["dateline","progname","id","testdat","targval"]


##  Main program :
##
def main():
    "Main line."
    
    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time() ; clockon = time.clock()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()

    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if not paradat.outpath.endswith(os.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
        if not os.path.isdir(paradat.outpath):  ##  try another outpath
            paradat.outpath = os.getcwd() + os.sep + "op" + os.sep
            if not os.path.isdir(paradat.outpath):
                os.makedirs(paradat.outpath)  ##  create o/p folder
            paradat.outfile = paradat.outpath + os.path.basename(paradat.outfile)

    ##  locate & read input files :
    tf = ''
    if paradat.testdat and os.path.isfile(paradat.testdat):
        print("data to be read from",paradat.testdat)
        r = getfile(paradat,paradat.testdat)
    else:
        tf = paradat.outpath + paradat.jobname + Dattail + ".dat"
        print("data to be sought on",tf)
        if os.path.isfile(tf):
            r = getfile(paradat,tf)
        else:
            raise RuntimeError("No test-data file found!")
    if r < 0:
        if not tf:
            tf = paradat.testdat
        raise RuntimeError("Problem reading test-data file "+tf)
    ##  vars 2b skipped will have already affected paradat.goodcols.
    paradat.testdat = os.path.abspath(tf)  ##  record name if successful.
    ##  next open prog file :
    if paradat.progfile:
        rf = paradat.progfile
    else:
        rf = paradat.outpath + paradat.jobname + Progtail  ##  always .py
    if rf.endswith(".py") and os.path.isfile(rf):
        paradat.progfile = open(rf,'r',encoding="utf-8")
        pylines = paradat.progfile.readlines()
        ##  used in o/p listing.
        paradat.progfile.close()
    else:
        print("Can't run without Python progfile (from pear.py) !")
        raise RuntimeError("Unsuitable progfile input :"+rf)

    ##  print(paradat.datamat[0])
    ##  print(paradat.datamat[-1])

    ##  prepare o/p files :
    if paradat.listfile:
        listname = paradat.listfile
    else:
        listname = paradat.outpath + paradat.jobname + Listtail + ".txt"
    paradat.listfile = open(listname,'w',encoding="utf8")
    if paradat.dumpfile:
        df = paradat.dumpfile
    else:
        df = paradat.outpath + paradat.jobname + Dumptail + ".dat"
    paradat.dumpfile = open(df,'w',encoding="utf-8")
    
    ##  indirectly import module containing py functions :
    head,tail = os.path.split(rf)
    print(head,tail)
    sys.path.append(head)
    plumsoft = loadlib(tail[:-3])  ##  no nice error-message if it fails
    print("Regression function successfully loaded from library :")
    print(plumsoft)
    print("[Expected to contain definitions beag_stabprep() & beag_decrule().]")
    ##  if this doesn't work, might as well fail early.
    ##  stab = plumsoft.beag_stabprep()
    ##  print(stab)

    ##  establish target values :
    if paradat.targval:
        targets(paradat)
        ##  should make .targvals & update .cols2use
    else:
        raise RuntimeError("No target expression (targval) given!!")
    tvalues = sorted(set(paradat.targvals))

    paradat.goodlist = sorted(paradat.goodcols)
    ##  print(paradat.goodlist)

    ##  identify categories :
    catlist = []
    catfrex = {}
    targvals = []  ##  will need target s.d.
    for r in range(paradat.rows):
        ##  tv = paradat.datamat[r][paradat.targcol]
        tv = paradat.targvals[r]
        targvals.append(tv)
        if tv in catfrex:
            catfrex[tv] += 1
        else:
            catfrex[tv] = 1  ## first encounter
        if not (tv in catlist):
            catlist.append(tv)
    paradat.catlist = sorted(catlist)
    paradat.cats = len(paradat.catlist)
    print("categories =",paradat.cats) ##  print(paradat.catlist)
    paradat.catfrex = catfrex

    paradat.targstat = bastatz(targvals)
    ##  next extract info, including catlist, from pylines :
    catlist = paradat.catlist
    for pl in pylines:
        regline = pl.rstrip()
        if "catlist =" in regline:
            bits = regline.split('=')
            catlist = eval(bits[1])
        elif "priorvec =" in regline:
            bits = regline.split('=')
            priorvec = eval(bits[1])
        elif "rule mode is" in regline:
            rulemode = "standard"
            if regline.endswith("tabular."):
                rulemode = "tabular"
        ##  training should should be third item in priorvec.
        ##  could be extracted ?!
    ##  print(catlist)
    cats = len(catlist)
    if rulemode != paradat.rulemode:
        print("Warning: rulemode being used is",rulemode)
    paradat.rulecats = catlist
    paradat.priorvec = priorvec
    paradat.rulemode = rulemode

    ##  prepare test phase :
    testnum = paradat.rows
    print(testnum)

    ##  test ruleset on test cases :
    allrows = list(range(paradat.rows))
    results = regtest(paradat,plumsoft,allrows,"testmode")
    nres = len(results)
    print("instances processed =",nres)
    paradat.nres = nres
    ##  summarize results :
    dustbins = hh = hits = 0
    icol = -1
    head = ["safeness","cellsize","id","stabcell","predval","trueval","mark"]
    head = Htab.join(head)  ##  R-style headline
    print(head,file=paradat.dumpfile)  ##  tab-delimited header
    if paradat.idfield and (paradat.idfield in paradat.taglist):
        icol = paradat.taglist.index(paradat.idfield)
    for i,tr in enumerate(sorted(results,reverse=True)):
        dumpline(paradat,tr,icol)
        if tr[-1] == '+':
            hits += 1
        if tr[1] < 10:  ##  small cellsize
            dustbins += 1
        else:
            if tr[-1] == '+':
                hh += 1
    hitrate = 100*hits/nres
    print("overall hitrate % =",round(hitrate,4))
    dustrate = 100*dustbins/nres
    if paradat.rulemode == "tabular":
        print("percentage small-sample decisions =",round(dustrate,4))
        hr = 100*hh / max(1,nres-dustbins)
        print("hitrate % omitting small-samples =",round(hr,4))
    paradat.dumpfile.close()

    ##  o/p of results :
    nout = showregs(paradat,results,pylines,"holdout")
    print("cases processed =",nout)
    print(NL,file=paradat.listfile)  ##  insert blanks
    parashow(paradat,paradat.listfile)
    listname = paradat.listfile.name
    print("main o/p listing written onto",os.path.abspath(listname))
    paradat.listfile.close()
    ##  parashow(paradat,fo) ; fo.close()
    
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff
    pybyebye()
    
    return  0




##  LOADLIB :
def loadlib (libname):
    "tries to import a library named at run time, using local namespace."

    locdict = {"grabbed" : '', "libname" : libname}
    impline = "import " + libname + " as grabbed"
    exec(impline,globals(),locdict)
    ##  print(':',locdict["grabbed"])

    return  locdict["grabbed"]
    ##  should avoid altering global namespace.


##  SHOWREGS :
def showregs (paradat,results,reglines,h):
    "writes eval info about results (assumed already sorted) onto file."
    ##  version for regression rule mode.
    
    ##  first some parameter info :
    for po in Outpars:
        pv = getattr(paradat,po,'')
        if pv:
            print(po.ljust(10),pv,file=paradat.listfile)
    print("\n===="+h+" trial :\n",file=paradat.listfile)
    
    ##  next headings for subsampling results :
    h1 = "rank  safeness case  #id#"
    h2 = "pred:true        cellsize    abdsiff    diffsqrd"
    vo = ''
    if paradat.idfield and (paradat.idfield in paradat.taglist):
        pi = paradat.taglist.index(paradat.idfield)
        vo = paradat.idfield[:Longname]
        h1 = h1.replace("#id#",vo)
    n = len(h1)
    h = h1.ljust(51) + h2
    print(h,file=paradat.listfile)

    ##  testmode results :
    marklist = []
    results.sort(reverse=True)
    rightvec = [] ; wrongvec = []  ##  right/wrong vecs for Hedges's g
    predvec = [] ; truevec = []  ##  for correlations
    svec = [] ; abserror = []
    tot1 = tot2 = 0.0
    nres = len(results)
    for j,r in enumerate(results):
        strength = r[0] ; svec.append(strength)
        cellsize = r[1]
        ix = r[2]
        cellcode = r[3]
        ##  minimal checking !
        predval,trueval = r[-2]
        predvec.append(predval)
        truevec.append(trueval)
        mark = r[-1]
        if mark == Goodsign:  ##  mark already computed
            rightvec.append(strength)
        else:
            ##  mark = Poorsign
            wrongvec.append(strength)
        d = abs(predval-trueval) ; abserror.append(d)
        tot1 += d ; tot2 += d*d
        predval = round(predval,4)  ## save space (better format?)  
        marklist.append(mark)
        rowvals = paradat.datamat[ix]
        if vo:
            h = str(rowvals[pi])  ##  case label
        else:
            h = str(ix)  ##  just use serial position
        h = h[:Longlab].ljust(Longlab)  ##  case label
        ##  print(j,strength,h,predval,mark,trueval,cellsize,predvals,
              ##  file=paradat.listfile)
        if j < 0:  ##  checking
            print(r)
            print([j+1,strength,ix,h,predval,mark,trueval,cellsize,[d,d*d]])
        showline(paradat,j+1,strength,ix,h,predval,mark,trueval,cellsize,
                 [d,d*d])

    marx = ''.join(marklist)
    print(marx,file=paradat.listfile)

    sr = len(rightvec) / max(1,nres)
    print("\n'success' percentage =",round(sr*100.0,2),
          file=paradat.listfile)
    cs = corsums(predvec,truevec)
    c = corrval(cs)
    print("pearson correlation between predicted & true vals =",
          round(c,4),file=paradat.listfile)
    cs = corsums(ranx(predvec),ranx(truevec))
    c = corrval(cs)
    print("spearman rank-correlation between predicted & true vals =",
          round(c,4),file=paradat.listfile)
    print("\nmean abs.error =",round(tot1/nres,4),file=paradat.listfile)
    print("mean error ^ 2 =",round(tot2/nres,4),file=paradat.listfile)

    ##  Hedges's g removed.
    s = "safeness"
    cs = corsums(abserror,svec) ; c = corrval(cs)
    print("correlation between",s,"& abs.error (negative better) = ",
          end='',file=paradat.listfile)
    print(round(c,4),file=paradat.listfile)
    ##  print(sum(rightvec)/len(rightvec))
    ##  print(sum(wrongvec)/len(wrongvec))

    ##  also show (intelligible?) version of regression rule :
    print("\nResultant rule from all training cases :",file=paradat.listfile)
    for outline in reglines:
        print(outline.rstrip(),file=paradat.listfile)

    return  len(results)


##  SHOWLINE :
def showline (paradat,rank,strength,ix,caselab,predval,mark,trueval,cellsize,
              predvals):
    "displays single line on paradat.listfile."

    fm = "{0:4d}  {1:8.2f} {2:4d}  {3:22s} {4:>10s} ~ {5:10s} {6:5d}"
    fm = fm.replace('~',mark,1)
    pv = str(predval)[:10] ; tv = str(trueval)[:10]
    out1 = fm.format(rank,strength,ix,caselab,pv,tv,cellsize)
    out2 = ''
    w = 12
    for pv in predvals:
        out2 = out2 + " {0:{1}.4f}".format(pv,w)
    print(out1+out2,file=paradat.listfile)


##  OUTSHOW :
def outshow (paradat,cmat,seen,fo):
    "shows confusion matrix."

    ##  display indicators :
    print("\ncases =",rc,file=fo)
    print("cases with unseen category labels =",rc-seen,file=fo)
    if seen > 0:
        if seen == 1:
            word = "case"
        else:
            word = "cases"
        if seen < rc:
            print("cases with known category labels =",seen,file=fo)
            print("[results below only apply to",seen,word+']',file=fo)
        print("hits = ",hits,file=fo)
        percent = 100 * hits / seen
        print("percent hits =",round(percent,2),file=fo)
        
    return


##  REGTEST :
def regtest (paradat,plumsoft,testsubs,testmode):
    "applies regression rule to (potentially) unseen test cases."

    reslist = []
    stab = plumsoft.runster_stabprep()
    tbar = paradat.priorvec[1]
    sd = paradat.priorvec[2]
    ##  priorvec used to pass over training mean & sd.
    
    for j in testsubs:
        rowvals = paradat.datamat[j]
        ##  stabpos,cellfrex = evalrule(paradat,thisrule,rowvals)
        varsack = Sack()  ##  unpack vals from list into Sack()
        ##  just create once ?
        for i,v in enumerate(rowvals):
            vn = paradat.taglist[i]
            setattr(varsack,vn,v)
        regvals = plumsoft.runster_regrule(varsack,stab)
        cellcode,predval,standev,smalldif,cellsize = regvals
        trueval = paradat.targvals[j]
        ##  mark = Goodsign
        ##  absdiff = abs(predval-trueval)
        ##  if not absdiff < smalldif:
            ##  mark = Poorsign
        mark = regmark(paradat,predval,trueval,paradat.rulecats)
        if paradat.rulemode != "tabular":
            strength = sd / (sd + abs(predval-tbar))
            ##  absdiff / smalldif  ??
        else:  ##  standard
            strength = sd / (sd + standev)
        reslist.append([strength,cellsize,j,cellcode,
                        (predval,trueval),mark])
        ##  print(reslist[-1])

    return  reslist


##  REGMARK :
def regmark (paradat,yhat,yval,fivenums):
    "uses Tukey five-nums to decide 'success' of prediction."

    k = khat = 0
    for y in fivenums[:2]:
        if yhat >= y:  k += 1
        if yval >= y:  khat += 1
    for y in fivenums[2:]:
        if yhat > y:  k += 1
        if yval > y:  khat += 1

    if k == khat:
        return  Goodsign

    return  Poorsign  ##  ybar should be in fivenums if len(fivenums) < 5.


##  GETFILE :
def getfile (paradat,fi):
    "obtains data grid from file."

    paradat.datamat,paradat.taglist,paradat.tagdict = matread(fi,paradat)
    paradat.rows = len(paradat.datamat)
    print("data rows =",paradat.rows)
    paradat.cols = len(paradat.taglist)
    print("data cols =",paradat.cols)
    print("column names :")
    print(paradat.taglist)
    if paradat.rows < Mincases:  ##  not enough data
        print("Not enough data cases !")
        print("Program halts.")
        return  -1
    goodcols = cols2use(paradat)
    paradat.goodcols = goodcols
    paradat.goodlist = sorted(paradat.goodcols)
    if len(goodcols) < 1:
        print("No numeric variables found in ",paradat.datfile,'!')
        print("Program halts.")
        return  -2

    return  0  ##  success, paradat altered




##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
