"Python3 library to handle rectangular data grids (as from R)."

##  GRIDLIB :
##  Python3 data-grid routines :
##  Copyright 2012-2016, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 15/09/2012
##  last revision : 12/09/2016


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)
    
import  math, os, string, sys
##  import  time, random


if sys.platform.startswith("linu"):  ##  presume linux
    sys.path.append("~/py3libs")
else:  ##  presume windows
    sys.path.append("c:\\py3libs")
##  in case executed from somewhere else.


##  local libraries :
from globdefs import *
from glib import  numeric, numerize


##  COLS2USE :
def cols2use (paradat,numsonly=True):
    "works out which columns to use in calculations."

    goodcols = set(range(paradat.cols))
    ##  lowtags = [t.lower() for t in paradat.taglist]
    ##  first remove non-numeric columns :
    for vn,varname in enumerate(paradat.taglist):
        nn = paradat.tagdict[varname]
        if numsonly and (nn < paradat.rows):
            goodcols.discard(vn)
    ##  don't remove target column at this stage :
    ##  goodcols.discard(paradat.targcol)
    ##  also remove vars that user asks to omit :
    if paradat.skipvars:
        bits = paradat.skipvars.split(',')  ##  comma-delimited
        for b in bits:
            bb = b.strip()
            if bb in paradat.taglist:
                bcol = paradat.taglist.index(bb)
                goodcols.discard(bcol)
            else:
                print("Warning : skipvar",b,"not found in taglist!")

    return  goodcols  ##  set of usable numeric cols.


##  MATDUMP :
def matdump (fo,dmat):
    "dumps numeric square matrix onto o/p file."

    n = len(dmat)
    lo = 0
    for dvec in dmat:
        lout = vecline(dvec)
        fo.write(lout + NL)
        lo += 1
    assert  lo == n
    
    return  n


##  VECLINE :
def vecline (datavec,decimals=5,dlim=Htab):
    "returns a delimited string from a list of numerics."

    lout = ''
    for v in datavec:
        if type(v) == type(9):
            sv = str(v)
        elif type(v) == type(9.9):
            sv = str(round(v,decimals))
        else:  ##  something wrong ?
            sv = str(v)
        lout = lout + sv + dlim

    return  lout[:-1]  ##  remove trailing delimiter


##  MATREAD :
def matread (filename,paradat,tabchar=Htab):
    "reads data grid with header as first line."

    fi = open(filename,'r',encoding="utf-8")
    ##  crash if failure!
    datlines = fi.readlines()
    h = datlines[0].strip()  ##  presumes header line first
    lc = len(datlines)
    taglist = h.split(tabchar)  ##  assumes tab as delimiter
    cols = len(taglist)
    print(lc,cols)
    ##  print(taglist)
    ##  make dictionary 4 guessing column types :
    tagdict = {}
    for t in taglist:
        tagdict[t] = 0

    datamat = []
    for r in range(1,lc):
        datlist = datlines[r].strip().split(tabchar)
        vals = len(datlist)
        if vals < 1:
            continue  ##  skip blank lines
        if vals < cols:
            print("Warning: dodgy data line :")
            print("items =",vals,"vals = ",datlist)
            return  datamat, taglist  ##  will force early exit
        rowdata = []
        for tp,item in enumerate(datlist):
            v = numerize(item)  ##  most values expected 2b numeric
            if numeric(v):
                rowdata.append(v)
                tagdict[taglist[tp]] += 1
            else:
                rowdata.append(item)  ##  probably string value
                tagdict[taglist[tp]] -= 1
        datamat.append(rowdata)
        
    return  (datamat,taglist,tagdict)


##  MAKEFULL :
def makefull (prepath,filename):
    "joins prepath to filename, with separator if missing."

    if prepath.endswith(os.sep):
        fullname = prepath + filename
    else:  ##  insert separator
        fullname = prepath + os.sep + filename

    return  fullname


##  VAR2COL :
def var2col (varname,varnames,goodcols=set()):
    "tries to find varname in varnames, only in goodcols if given."

    if not goodcols:
        goodcols = set(range(len(varnames)))
    if varname in varnames:
        vpos = varnames.index(varname)
        if vpos in goodcols:
            return  vpos  ##  position of varname

    return  -1  ##  failure signal


##  BASTATZ :
def bastatz (numlist):
    "basic stats routine for a list of numbers."

    fn = float(len(numlist))
    sum1 = 0.0
    sum2 = 0.0
    for x in numlist:
        sum1 += x
        sum2 += x**2

    statsack = Sack(0)
    ##  defaults for dodgy inputs :
    statsack.mean = sum1
    statsack.var1 = statsack.var2 = 0.0
    statsack.std1 = statsack.std2 = 0.0
    if fn > 0.0:
        statsack.mean = sum1 / fn
        statsack.var1 = sum2 / fn - statsack.mean**2
        ##  print "$$",statsack.var1
        statsack.std1 = math.sqrt(max(0.0,statsack.var1))
        if fn > 1.0:
            statsack.var2 = statsack.var1 * (fn/(fn-1.0))
            statsack.std2 = math.sqrt(max(0.0,statsack.var2))

    return statsack


##  MIDVALUE :
def midvalue (xvec):
    "simple median."

    vecsize = len(xvec)
    if vecsize < 1:
        return  0.0  ##  None better ?
    if vecsize % 2 == 1:  ##  uneven
        m = vecsize // 2
        return  sorted(xvec)[m]
    else:  ##  even
        m = (vecsize-1)//2
        v1,v2 = sorted(xvec)[m:m+2]
        return  (v1+v2)*0.5


##  FIVENUM :
def fivenum (xvec):
    "Tukey-style five-number summary of numeric vector."

    n1 = len(xvec) - 1
    xs = sorted(xvec)
    if n1 < 5:
        return xs
    v5 = [xs[0]]  ##  minimum
    for f in [0.25,0.5,0.75]:
        lo = int(n1*f)
        hi = n1 - int((1.0-f)*n1)
        v5.append((xs[lo]+xs[hi])*0.5)
    v5.append(xs[-1])  ##  maximum

    return  v5



##  VECDIST :
def vecdist (xvec,yvec,p=2):
    "computes Euclidean or p-distance with 1 sweep, Euclid as default."

    n1 = len(xvec) ; n2 = len(yvec)
    n = min(n1,n2)
    ##  what if n1 != n2 ?
    if n < 1:
        return  0.0
    dsum = 0.0
    for j in range(n):
        d = abs(xvec[j]-yvec[j])
        dsum += d**p  ##  p==1 for city-block

    return  dsum ** (1/p)  ##  could divide by n outside ?


##  DISTEST :
def distest (grid):
    "tests whether data grid looks like a distance/dissimilarity matrix."

    rows = len(grid)
    if rows > 0:
        cols = len(grid[0])
    else:
        return  False  ##  could null grid count as sq matrix ??
    if rows != cols:
        return  False
    ##  size okay, test contents :
    for r in range(rows):
        if grid[r][r] != 0.0:
            return  False  ##  must have zero diagonal
        lo = min(grid[r])
        if lo < 0.0:  ##  cd subzero dissims make sense ??
            return  False  ##  negative distances disallowed at present

    return  True
    ##  allows dissimilarity matrices, where i,j locs != j,i locs.


##  COPYGRID :
def copygrid (grid):
    "returns a copy of given data grid."

    newgrid = []
    for r in grid:
        rowdata = []
        for v in r:
            rowdata.append(v)
        newgrid.append(rowdata)

    return  newgrid


##  CENTDIST :
def centdist (paradat,mvec,gvec):
    "computes distance of grid item from midclust."

    xvec = [gvec[gc] for gc in paradat.goodlist]
    if paradat.scaling > 0:
        xvec = rescale(paradat,xvec)
    assert len(xvec) == len(mvec)
    p = 2
    if paradat.distmode == "city":
        p = 1
    dx = vecdist(xvec,mvec,p)

    return  dx


##  RESCALE :
def rescale (xvec,scaling):
    "rescales a numeric vector."

    ##  xvec must be numeric (right thru!)
    xmin = min(xvec) ; xmax = max(xvec)
    xr = (xmax - xmin) or 1.0
    bs = bastatz(xvec)
    vx = []
    for x in xvec:
        if scaling < 1:  ##  unaltered
            vx.append(x)
        elif paradat.scaling == 1:  ##  scaled by range
            xx = (x - xmin) / xr
            vx.append(xx)
        else:  ##  scaled as zedscore
            xx = x - bs.mean
            if bs.std2 > 1e-12:
                xx = xx / bs.std2
            vx.append(xx)

    return  vx


##  PRECREC :
def precrec (cmat):
    "returns precisions & recalls of confusion matrix held as freqtab of pairs."

    c1 = set() ; c2 = set()
    pairlist = sorted(cmat.keys())  ##  order in pair: predcat,truecat
    for k1,k2 in pairlist:
        c1.add(k1) ; c2.add(k2)
    c1 = sorted(c1) ; c2 = sorted(c2)
    ptab = {} ; rtab = {}
    ##  obtain totals :
    for p in pairlist:
        f = cmat[p]
        pn = ptab.get(p[0],0)
        ptab[p[0]] = pn + f
        rn = rtab.get(p[1],0)
        rtab[p[1]] = rn + f
    ##  print(ptab,rtab)
    ##  compute percentages :
    for c in c1:
        ok = cmat.get((c,c),0)
        ptab[c] = 100.0 * ok / ptab.get(c,1)  ##  precision in class c
    for c in c2:
        ok = cmat.get((c,c),0)
        rtab[c] = 100.0 * ok / rtab.get(c,1)  ##  recall of class c

    return  (ptab,rtab)


##  MATSHOW :
def matshow (fo,cmat,longest=20):
    "write confusion matrix (dict of pairs) to already open file."

    fo.write(NL + "Confusion matrix :" + NL + NL)
    ##  make local category list (could include unknown cats in holdout mode) :
    cats = set()
    predcats = set()
    for k1,k2 in cmat.keys():
        cats.add(k1) ; cats.add(k2)
        predcats.add(k1)
    catlist = sorted(cats)
    ##  find widest class name :
    w = 4
    for c in catlist:
        wx = len(str(c))
        if wx > w:
            w = wx
    w = min(longest,w)
    lout = "Truecat =".ljust(10+w)
    for c in catlist:
        lout = lout + str(c)[:longest].rjust(w+1)
    fo.write(lout + NL)
    for pred in sorted(predcats):
        lout = "Predcat : " + str(pred).ljust(w)
        for truecat in catlist:
            pair = (pred,truecat)
            f = cmat.get(pair,0)
            lout = lout + " {0:{1}d}".format(f,w)
        fo.write(lout + NL)

    fo.write(NL)
    return


##  SCKAPPA2 :
def sckappa2 (ctab,codelist):
    "Siegel & Catellan's version of Kappa, for 2-way confusion matrix held as table."

    nc = len(codelist)
    chap = 0.0  ##  proportion agreement expected by chance
    totf = 0.0  ##  total frequency, all cells in table
    hits = 0
    codefrex = {}
    for k1 in codelist:
        ##  codelist holds all valid category codes.
        for k2 in codelist:
            pair = (str(k1), str(k2))  ##  numerals & numbers treated as equal
            f = ctab.get(pair,0)
            if k1 == k2:
                hits += f
            f1 = codefrex.get(k1,0)
            codefrex[k1] = f1 + f
            f2 = codefrex.get(k2,0)
            codefrex[k2] = f2 + f
            totf = totf + f
    totf = max(Tiny,totf)  ##  to prevent zero division
    for k in codelist:
        f = codefrex.get(k,0)
        pk = f / (2.0 * totf)
        chap += (pk * pk)
        ##  print k,f,pk,chap
    ##  chap is sum of squared probs.
    pa = hits / totf
    ##  print  hits, pa, totf
    k = (pa - chap) / max((1 - chap),Tiny)
    return  (k, chap,hits,pa,totf)


##  HEDGESG :
def hedgesg (x1,x2):
    "effect-size measure for 2 independent groups: Cohen's d with pooled variance."

    n1 = len(x1) ; n2 = len(x2)
    if (n1+n2) < 4 or n1 < 1 or n2 < 1:
        return  0.0
    s1 = bastatz(x1)
    s2 = bastatz(x2)
    diff = s1.mean - s2.mean
    if (n1+n2) <= 4:  ##  puny sample
        return  diff / max(s1.std2,s2.std2)
    else:  ##  normal case
        pooledv = ((n1-1)*s1.var2 + (n2-1)*s2.var2) / (n1+n2-2)
        return  diff / math.sqrt(pooledv)


##  CORSUMS :
def corsums (x,y):
    "produces sums required for correlation coefficient, in a Sack()."

    cs = Sack(0)
    n = len(x)
    n = min(n,len(y))  ##  chops extra if unequal
    if n < 1:
        print("Problem in corsums!!")
        return cs
    sx = sy = 0.0
    sxsq = sysq = 0.0
    sumprods = 0.0
    for i in range(n):
        sx += x[i] ; sy += y[i]
        sxsq += x[i]**2 ; sysq += y[i]**2
        sumprods += x[i] * y[i]

    cs.n = float(n)
    cs.sumprods = sumprods
    cs.sx = sx
    cs.sy = sy
    cs.sxsq = sxsq
    cs.sysq = sysq
    return  cs


##  CORRVAL :
def corrval (cs,epsilon=1e-23):
    "computes correlation coefficient from sums etc. in Sack()."

    cnum = cs.sumprods - (cs.sx*cs.sy) / cs.n
    tx = cs.sxsq - cs.sx**2 / cs.n
    ty = cs.sysq - cs.sy**2 / cs.n
    ##  try:
        ##  dnom = math.sqrt(tx*ty)
    ##  except ValueError as xxxx:
        ##  print(xxxx)
        ##  print(tx,ty,cnum,cs.n)
        ##  print(cs.sx,cs.sy,cs.sxsq,cs.sysq)
        ##  dnom = 0
    ##  dnom = math.sqrt(abs(tx*ty))  ##  danger of rounding errors
    dnom = math.sqrt(max(0.0,tx*ty))
    if dnom > epsilon:
        return  cnum / dnom
    elif tx < epsilon and ty < epsilon:
        return  1.0  ##  both flat
    else:
        return  0.0  ##  null case
    ##  semi-safe, but vulnerable to rounding errors.


##  RANX :
def ranx (xvec):
    "returns rank vector from raw data vector, with ties given mean positions."

    nx = len(xvec)
    ##  xvec.sort()  ##  ascending
    vtab = {}
    for x in xvec:
        fx = vtab.get(x,0)
        vtab[x] = fx + 1
    vvec = list(vtab.items())
    vvec.sort()
    ranktab = {}
    p = 0
    ranksum = 0
    for v in vvec:
        vx = v[0] ; vf = v[1]
        rx = (ranksum*2+vf)*0.5-0.5  ##  plus 0.5 if lowest rank is 1
        ranksum += vf
        ##  print(p,vx,vf,rx)
        ranktab[vx] = rx
        p += 1

    rankvec = []
    for x in xvec:
        rankvec.append(ranktab[x])
    return  rankvec  ##  returning ranktab more useful ?


##  MADM :
def madm (xvec):
    "returns median of absolute difs from median."

    m = midvalue(xvec)
    n = len(xvec)
    dabs = [abs(m-xvec[j]) for j in range(n)]
    ##  print(sorted(dabs))
    return  midvalue(dabs)


##  quick tests :

if 0 > 0:
    z = [1,2,3,4,0,9,9,7,0,Gold]
    print(sorted(z))
    print(midvalue(z))
    print(madm(z))
    bs = bastatz(z)
    print(bs.mean,bs.std1,bs.std2)
