"Python3 BEAGLE rule-evaluation tester."

##  LEAFLET :
##
##  Likelihood Estimator And Forecaster's Logical Evaluation Tester in Python3.
##  Copyright (c), 2016, R.S. Forsyth.
##  ancestral progenitor : Sept, 1980.
##  first version : 17/12/2015
##  last revision : 08/08/2016


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  copy
import  math
import  os, sys
import  random, time

##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  parashow
from glib import  showobj
from glib import  openfile
from glib import  pybyebye
from glib import  request
from gridlib import  cols2use, matread
from gridlib import  matdump, vecline
from gridlib import  var2col
from gridlib import  bastatz, midvalue
from gridlib import  vecdist
from gridlib import  copygrid, distest
from gridlib import  matshow, precrec
from gridlib import  sckappa2
from gridlib import  hedgesg
from herb import  windist
import  rulelib
import  treelib


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ##  ("atomize",1,0,1),
    ("randseed",0,0,999999999),  ##  random seed, <2 means haphazard seed
    ("jobname","beagle16"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ("testdat",''),  ##  file for input of data
    ##  ("cols2use",''),  ##  variables to be used
    ("targval",''),  ##  target variable
    ("targcol",-1),  ##  target column number
    ("idfield",''),  ##  varname for o/p label
    ("skipvars",''),  ##  variables to be excluded [inoperative?]
    ("listfile",''),  ##  main output file
    ("dumpfile",''),  ##  output file for subsampling results (?)
    ("progfile",''),  ##  file containing Python code
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"leaf.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "Likelihood Estimator And Forecaster's Language Export Tester v3.2"
##  Gold = 5.0**0.5*0.5+0.5
Mincases = 2
Dattail = "_dat2"
Dumptail = "_pout"
Listtail = "_prop"  ##  program o/p
Progtail = "_prog.py"
Maxcats = 64
Dustbin = "??!!"
Goodsign = '+'
Poorsign = '-'
Longname = 26  ##  length after which category names will be truncated
Longlab = 20  ##  length after which label fields will be truncated
Outpars = ["dateline","progname","id","testdat","targval"]


##  Main program :
##
def main():
    "Main line."
    
    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time() ; clockon = time.clock()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()

    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if not paradat.outpath.endswith(os.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
        if not os.path.isdir(paradat.outpath):  ##  try another outpath
            paradat.outpath = os.getcwd() + os.sep + "op" + os.sep
            if not os.path.isdir(paradat.outpath):
                os.makedirs(paradat.outpath)  ##  create o/p folder
            paradat.outfile = paradat.outpath + os.path.basename(paradat.outfile)

    ##  prepare o/p files :
    if paradat.listfile:
        listname = paradat.listfile
    else:
        listname = paradat.outpath + paradat.jobname + Listtail + ".txt"
    paradat.listfile = open(listname,'w',encoding="utf8")

    if paradat.dumpfile:
        df = paradat.dumpfile
    else:
        df = paradat.outpath + paradat.jobname + Dumptail + ".txt"
    paradat.dumpfile = open(df,'w',encoding="utf-8")

    ##  locate & read input files :
    tf = ''
    if paradat.testdat and os.path.isfile(paradat.testdat):
        print("data to be read from",paradat.testdat)
        r = getfile(paradat,paradat.testdat)
    else:
        tf = paradat.outpath + paradat.jobname + Dattail + ".dat"
        print("data to be sought on",tf)
        if os.path.isfile(tf):
            r = getfile(paradat,tf)
        else:
            raise RuntimeError("No test-data file found!")
    if r < 0:
        if not tf:
            tf = paradat.testdat
        raise RuntimeError("Problem reading test-data file "+tf)
    ##  vars 2b skipped will have already affected paradat.goodcols.
    paradat.testdat = os.path.abspath(tf)  ##  record name if successful.
    ##  next open prog file :
    if paradat.progfile:
        rf = paradat.progfile
    else:
        rf = paradat.outpath + paradat.jobname + Progtail  ##  always .py
    if rf.endswith(".py") and os.path.isfile(rf):
        paradat.progfile = open(rf,'r',encoding="utf-8")
        pylines = paradat.progfile.readlines()
        ##  used in o/p listing.
        paradat.progfile.close()
    else:
        print("Can't run without Python progfile (from plum.py) !")
        raise RuntimeError("Unsuitable progfile input :"+rf)

    print(paradat.datamat[0])
    print(paradat.datamat[-1])

    ##  indirectly import module containing py functions :
    head,tail = os.path.split(rf)
    print(head,tail)
    sys.path.append(head)
    plumsoft = loadlib(tail[:-3])  ##  no nice error-message if it fails
    print("Decision function successfully loaded from library :")
    print(plumsoft)
    print("[Expected to contain definitions beag_stabprep() & beag_decrule().]")
    ##  if this doesn't work, might as well fail early.
    ##  stab = plumsoft.beag_stabprep()
    ##  print(stab)

    ##  establish target values :
    for p in pylines:
        datline = p.strip()
        if "##  target : " in datline:
            bits = datline.split(" : ",1)
            pytarget = bits[1]
            ##  print(bits)
    print("Para/rule target exps :",paradat.targval,pytarget,file=paradat.dumpfile)
    if pytarget != paradat.targval:
        print("Warning: target differs from parameter file to code file!")
        print("was :",paradat.targval) ; print("is :",pytarget)
        paradat.targval = pytarget

    if paradat.targval:
        targets(paradat)
        ##  should make .targvals & update .cols2use
    else:
        raise RuntimeError("No target expression (targval) given!!")
    
    ##  identify categories :
    catlist = []
    catfrex = {}
    for r in range(paradat.rows):
        ##  tv = paradat.datamat[r][paradat.targcol]
        tv = paradat.targvals[r]
        if tv in catfrex:
            catfrex[tv] += 1
        else:
            catfrex[tv] = 1  ## first encounter
        if not (tv in catlist):
            catlist.append(tv)
    paradat.catlist = sorted(catlist)
    paradat.cats = len(paradat.catlist)
    print("input categories =",paradat.cats) ; print(paradat.catlist)
    paradat.catfrex = catfrex
    paradat.priorvec = preprobs(paradat,catfrex)  ##  adjusted prior probs

    ##  compatibility issues :
    ##  assert paradat.targval == bestrule.targval
    ##  paradat.priorvec = bestrule.priorvec
    ##  paradat.subrules = len(bestrule.testlist)
    ##  paradat.stabrows = 2 ** paradat.subrules
    ##  print(bestrule.catlist) ; print(paradat.catlist)
    ##  next extract catlist from declines :
    catlist = paradat.catlist
    for decline in pylines:
        if "catlist =" in decline:
            bits = decline.split('=')
            catlist = eval(bits[1])
            break
    print("rule categories :",catlist) ; cats = len(catlist)
    paradat.rulecats = catlist

    ##  prepare test phase :
    testnum = paradat.rows
    print(testnum)

    ##  test ruleset on test cases :
    allrows = list(range(paradat.rows))
    results = ruletest(paradat,plumsoft,allrows,"testmode")
    nres = len(results)
    print("instances processed =",nres)
    paradat.nres = nres
    ##  summarize results :
    dustbins = hh = hits = 0
    for tr in sorted(results,reverse=True):
        print(tr,file=paradat.dumpfile)
        if tr[-1] == '+':
            hits += 1
        if tr[1] < 10:  ##  small cellsize
            dustbins += 1
        else:
            if tr[-1] == '+':
                hh += 1
    hitrate = 100*hits/nres
    print("overall hitrate % =",round(hitrate,4))
    dustrate = 100*dustbins/nres
    print("percentage small-sample decisions =",round(dustrate,4))
    hr = 100*hh / max(1,nres-dustbins)
    print("hitrate % omitting small-samples =",round(hr,4))
    paradat.dumpfile.close()
    
    ##  o/p of results :
    nout = showres(paradat,results,pylines,"holdout")
    ##  print("cases processed =",nout)
    assert  nout == paradat.nres
    print(NL,file=paradat.listfile)  ##  insert blanks
    parashow(paradat,paradat.listfile)
    paradat.listfile.close()
    ##  parashow(paradat,fo) ; fo.close()
    
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff

    pybyebye()
    return  0




##  TARGETS :
def targets (paradat):
    "sets up .targvals from .targval expression & updates .cols2use."

    tx = paradat.targval.strip()
    paradat.goodlist = sorted(paradat.goodcols)
    allcols = set(range(paradat.cols))
    paradat.strcols = sorted(allcols - paradat.goodcols)
    paradat.cols2use = allcols
    paradat.targvals = []
    if tx in paradat.taglist:  ##  simple variable
        paradat.targcol = paradat.taglist.index(tx)
        for rowvals in paradat.datamat:
            paradat.targvals.append(str(rowvals[paradat.targcol]))
        paradat.cols2use.discard(paradat.targcol)
    else:  ##  presume expression
        print("Evaluating target expression : ",tx)
        ##  convert to internal tree format :
        toks = treelib.treetox(tx)
        print(toks)
        pole = treelib.tox2pole(toks)
        ##  print(pole)
        targtree = treelib.poletree(paradat,pole)
        for rowvals in paradat.datamat:
            tv = treelib.treeval(targtree,paradat,rowvals)
            paradat.targvals.append(str(tv))  ##  hold as string
        ##  remove variables in target expression :
        nodelist = treelib.getnodes(targtree,paradat)
        for n in nodelist:
            ##showobj(n)
            if n.type == "feature":
                paradat.cols2use.discard(n.item)
            elif n.type == "stringy":
                paradat.cols2use.discard(n.item[0])  ##  first part of pair is colnum

    print("Target values :",paradat.targvals,file=paradat.dumpfile)
    ##  print(paradat.goodcols)
    ##  print(paradat.strcols)
    print("Cols 2b used :",paradat.cols2use,file=paradat.dumpfile)
    ##  no need to update .cols2use ??
    
    return


##  LOADLIB :
def loadlib (libname):
    "tries to import a library named at run time, using local namespace."

    locdict = {"grabbed" : '', "libname" : libname}
    impline = "import " + libname + " as grabbed"
    exec(impline,globals(),locdict)
    ##  print(':',locdict["grabbed"])

    return  locdict["grabbed"]
    ##  should avoid altering global namespace.


##  PREPROBS :
def preprobs (paradat,catfrex):
    "goes from table of class frex to list of category priors."

    w = []
    for j,c in enumerate(paradat.catlist):  ##  standard order
        f = catfrex[c]
        w.append(f**0.5)  ##  proportional to sqrt of freq
    wtot = sum(w)

    return  [fx/wtot for fx in w]
    

##  SHOWRES :
def showres (paradat,results,declines,h):
    "writes eval info about results (assumed already sorted) onto file."

    ##  first some parameter info :
    for po in Outpars:
        pv = getattr(paradat,po,'')
        if pv:
            print(po.ljust(10),pv,file=paradat.listfile)
    print("\n===="+h+" trial :\n",file=paradat.listfile)
    
    ##  next headings for subsampling results :
    h1 = "rank  strength case  #id#"
    h2 = "pred:true      cellsize  predvals"
    vo = ''
    if paradat.idfield and (paradat.idfield in paradat.taglist):
        pi = paradat.taglist.index(paradat.idfield)
        vo = paradat.idfield[:Longname]
        h1 = h1.replace("#id#",vo)
    n = len(h1)
    h = h1.ljust(49) + h2
    print(h,file=paradat.listfile)

    ##  next extract catlist from declines :
    ##  catlist = paradat.catlist
    ##  for decline in declines:
        ##  if "catlist =" in decline:
            ##  bits = decline.split('=')
            ##  catlist = eval(bits[1])
            ##  break
    print(paradat.rulecats)
    catlist = paradat.rulecats
    cats = len(catlist)

    ##  testmode results :
    marklist = []
    results.sort(reverse=True)
    rightvec = [] ; wrongvec = []  ##  right/wrong vecs for Hedges's g
    for j,r in enumerate(results):
        strength = r[0]
        cellsize = r[1]
        ix = r[2]
        assert cellsize == sum(r[3])  ##  minimal checking !
        predvals = r[4]
        predval,trueval = r[-2]
        if predval == trueval:
            mark = Goodsign
            rightvec.append(strength)
        elif trueval in catlist:
            mark = Poorsign
            wrongvec.append(strength)
        else:
            mark = '?'
        marklist.append(mark)
        rowvals = paradat.datamat[ix]
        if vo:
            h = str(rowvals[pi])  ##  case label
        else:
            h = str(ix)  ##  just use serial position
        h = h[:Longlab].ljust(Longlab)  ##  case label
        ##  print(j,strength,h,predval,mark,trueval,cellsize,predvals,
              ##  file=paradat.listfile)
        showline(paradat,j+1,strength,ix,h,predval,mark,trueval,cellsize,
                 predvals)

    marx = ''.join(marklist)
    print(marx,file=paradat.listfile)
    
    cmat = {}  ##  empty confusion matrix
    seen = len(rightvec) + len(wrongvec)
    for r in results:
        pair = r[-2]
        ##  if pair[-1] in paradat.catlist:
            ##  seen += 1
        ##  print(r,pair)
        f = cmat.get(pair,0)
        cmat[pair] = f + 1

    outshow(paradat,cmat,seen,paradat.listfile)

    ##  also differential of right & wrong strengths :
    try:
        hg = hedgesg(rightvec,wrongvec)
    except (ValueError,ZeroDivisionError) as xxxx:
        hg = 0.0
    print("Hedges's g (z-gap) between strengths of right & wrong answers =",
          round(hg,4),file=paradat.listfile)
    ##  print(sum(rightvec)/len(rightvec))
    ##  print(sum(wrongvec)/len(wrongvec))

    ##  also show (intelligible?) version of decision rule :
    ##  rout = rule2str(paradat,decrule)
    print("\nRule used shown below, with target :",paradat.targval,
          file=paradat.listfile)
    print('~',catlist,NL,file=paradat.listfile)
    ##  print(rout,file=paradat.listfile)
    for lout in declines:
        print(lout.rstrip(),file=paradat.listfile)
    print(file=paradat.listfile)
    ##  showdeli(paradat,declist,paradat.listfile)

    return  len(results)


##  SHOWLINE :
def showline (paradat,rank,strength,ix,caselab,predval,mark,trueval,cellsize,
              predvals):
    "displays single line on paradat.listfile."

    fm = "{0:4d}  {1:8.2f} {2:4d}  {3:20s} {4:>10s} ~ {5:10s} {6:4d}"
    fm = fm.replace('~',mark,1)
    pv = str(predval)[:10] ; tv = str(trueval)[:10]
    out1 = fm.format(rank,strength,ix,caselab,pv,tv,cellsize)
    out2 = ''
    if len(predvals) > 7:
        w = 5
    else:
        w = 6
    for pv in predvals:
        out2 = out2 + " {0:{1}.2f}".format(pv,w)
    print(out1+out2,file=paradat.listfile)


##  OUTSHOW :
def outshow (paradat,cmat,seen,fo):
    "shows confusion matrix."
    
    ##  show confusion matrix :
    ptab,rtab = precrec(cmat)
    keylist = sorted(cmat.keys())  ##  pairs
    wide = max([len(str(k)) for k in keylist])
    wide = wide // 2
    rc = paradat.nres  ##  number of cases actually processed
    ##  matshow & precall expect pred,true order :
    matshow(fo,cmat,Longname)
    sk = sckappa2(cmat,paradat.catlist)  ##  keylist ??
    k2 = round(sk[0],4)
    ##  print(keylist,paradat.catlist)
    hits = sk[2]
    print("Kappa value = ",k2,file=fo)
    print("Precision (%) by category :",file=fo)
    for c in sorted(ptab.keys()):
        print(str(c)[:Longname].ljust(wide),
              round(ptab[c],4),file=fo)
    print("Recall (%) by category :",file=fo)
    for c in sorted(rtab.keys()):
        print(str(c)[:Longname].ljust(wide),
              round(rtab[c],4),file=fo)

    ##  display indicators :
    print("\ncases =",rc,file=fo)
    print("cases with unseen category labels =",rc-seen,file=fo)
    if seen > 0:
        if seen == 1:
            word = "case"
        else:
            word = "cases"
        if seen < rc:
            print("cases with known category labels =",seen,file=fo)
            print("[results below only apply to",seen,word+']',file=fo)
        print("hits = ",hits,file=fo)
        percent = 100 * hits / seen
        print("percent hits =",round(percent,2),file=fo)
        
    return


##  MODFRAX :
def modfrax (freqvec,digs=7):
    "attenuates frequencies to produce bayesian multipliers."

    n = sum(freqvec) ; c = len(freqvec)
    slug = Gold
    pvec = [f+slug/c for f in freqvec]
    psum = sum(pvec)

    return  [round(p/psum,digs) for p in pvec]


##  RULETEST :
def ruletest (paradat,plumsoft,testsubs,testmode):
    "applies rule on (potentially) unseen test cases."

    reslist = []
    stab = plumsoft.beag_stabprep()
    for j in testsubs:
        rowvals = paradat.datamat[j]
        ##  stabpos,cellfrex = evalrule(paradat,thisrule,rowvals)
        varsack = Sack()  ##  unpack vals from list into Sack()
        for i,v in enumerate(rowvals):
            vn = paradat.taglist[i]
            setattr(varsack,vn,v)
        rulevals = plumsoft.beag_decrule(varsack,stab)
        b,predcat,cellfrex,post = rulevals
        cellsize = sum(cellfrex)
        stabpos = b
        ##  compute class label here :
        ##  apply attentuation factor :
        ##  slug = Gold/paradat.cats  ##  sufficient ??
        ##  modfrex = [f+slug for f in cellfrex]
        ##  go posterior (Bayes++) :
        ##  post = [modfrex[j]*paradat.priorvec[j] for j in range(paradat.cats)]
        ##  psum = sum(post)
        post = [round(pv,7) for pv in post]  ##  compatible with leaf
        pc = max(post) ; pc = post.index(pc)
        pred = paradat.catlist[pc]
        ##  trueval = rowvals[paradat.targcol]
        trueval = paradat.targvals[j]
        mark = Goodsign
        if not (trueval in paradat.rulecats):
            mark = '?'
        elif predcat != trueval:
            mark = Poorsign
        strength = round(windist(post),7)
        reslist.append([strength,cellsize,j,cellfrex,post,stabpos,
                        (predcat,trueval),mark])
        ##  print(reslist[-1])

    return  reslist


##  GETFILE :
def getfile (paradat,fi):
    "obtains data grid from file."

    paradat.datamat,paradat.taglist,paradat.tagdict = matread(fi,paradat)
    paradat.rows = len(paradat.datamat)
    print("data rows =",paradat.rows)
    paradat.cols = len(paradat.taglist)
    print("data cols =",paradat.cols)
    print("column names :")
    print(paradat.taglist)
    if paradat.rows < Mincases:  ##  not enough data
        print("Not enough data cases !")
        print("Program halts.")
        return  -1
    goodcols = cols2use(paradat)
    paradat.goodcols = goodcols
    paradat.goodlist = sorted(paradat.goodcols)
    if len(goodcols) < 1:
        print("No numeric variables found in ",paradat.datfile,'!')
        print("Program halts.")
        return  -2

    return  0  ##  success, paradat altered




##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
