"Python3 BEAGLE program to convert rule into programming code (in Python or R)."

##  PLUM :
##
##  Programming Language Utility Module in Python3.
##  Copyright (c), 2016, R.S. Forsyth.
##  ancestral progenitor : July, 1985.
##  first version : 15/12/2015
##  last revision : 08/08/2016


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  copy
import  math
import  os, sys
import  random, time

##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  parashow
from glib import  showobj
from glib import  openfile
from glib import  pybyebye
from glib import  request
from gridlib import  cols2use, matread
from gridlib import  matdump, vecline
from gridlib import  var2col
from gridlib import  bastatz, midvalue
from gridlib import  vecdist
from gridlib import  copygrid, distest
from gridlib import  matshow, precrec
from gridlib import  sckappa2
from gridlib import  hedgesg
import  rulelib, treelib


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ##  ("atomize",1,0,1),
    ##  ("filetail",''),
    ("jobname","beagle16"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ##  ("testdat",''),  ##  file for input of data
    ##  ("cols2use",''),  ##  variables to be used
    ("targval",''),  ##  target variable
    ("targcol",-1),  ##  target column number
    ##  ("idfield",''),  ##  varname for o/p label
    ##  ("skipvars",''),  ##  variables to be excluded
    ("proglang",'r'),  ##  r is current default [alternative = py]
    ("listfile",''),  ##  main output file
    ("progfile",''),  ##  output file for subsampling results
    ("rulefile",''),  ##  file for o/p of rules
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"plum.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "Programming Language Utility Module v3.2"
##  Gold = 5.0**0.5*0.5+0.5
Mincases = 2
Progtail = "_prog"
Listtail = "_plum"
Ruletail = "_rule"
Maxcats = 64
Dustbin = "??!!"
Buck = '$'
Lbra = '('
Rbra = ')'


##  Main program :
##
def main():
    "Main line."
    
    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time() ; clockon = time.clock()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()

    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if not paradat.outpath.endswith(os.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
        if not os.path.isdir(paradat.outpath):  ##  try another outpath
            paradat.outpath = os.getcwd() + os.sep + "op" + os.sep
            if not os.path.isdir(paradat.outpath):
                os.makedirs(paradat.outpath)  ##  create o/p folder
            paradat.outfile = paradat.outpath + os.path.basename(paradat.outfile)

    ##  prepare o/p files :
    if not os.path.isdir(paradat.outpath):
        os.makedirs(paradat.outpath)
        print(paradat.outpath,"created.")
    if paradat.listfile:
        listname = paradat.listfile
    else:
        listname = paradat.outpath + paradat.jobname + Listtail + ".txt"
    paradat.listfile = open(listname,'w',encoding="utf8")
    if paradat.progfile:
        df = paradat.progfile
    else:
        df = paradat.outpath + paradat.jobname + Progtail + '.' + paradat.proglang
    paradat.progfile = open(df,'w',encoding="utf-8")

    ##  next open rule file :
    if paradat.rulefile:
        rf = paradat.rulefile
    else:
        rf = paradat.outpath + paradat.jobname + Ruletail + ".txt"
    paradat.rulefile = open(rf,'r',encoding="utf-8")
    ##  read from rule file :
    datlines = paradat.rulefile.readlines()
    for j in []:
        print(datlines[j])
    bits = datlines[0].split(':',1)
    print(bits)
    paradat.traindat = bits[1].strip()
    ##  print(os.path.isfile(paradat.traindat))
    if os.path.isfile(paradat.traindat):  ##  obtain taglist
        fi = open(paradat.traindat,'r',encoding="utf-8")
        tagline = fi.readline()  ##  top line gives tags
        paradat.taglist = tagline.split()
        print("tagnames:")
        print(paradat.taglist)  ##  later 2b passed in rulefile
    else:  ##  won't work without training file at present
        raise RuntimeError("Can't find training data file : "+paradat.traindat)
    mainrule = rulelib.getrule(paradat,datlines)
    paradat.rulefile.close()
    ##  showobj(mainrule)

    ##  deal with compatibility issues :
    assert paradat.targval == mainrule.targval
    paradat.priorvec = mainrule.priorvec
    paradat.rulemode = mainrule.rulemode
    paradat.subrules = len(mainrule.testlist)
    paradat.stabrows = mainrule.stabrows  ##  computed in getrule()
    paradat.cols = mainrule.cols
    assert paradat.cols == len(paradat.taglist)
    paradat.rows = mainrule.rows
    bout = rulelib.ruleshow(paradat,mainrule)
    ##  for b in bout:
        ##  print(b)

    ##  pull in rule skeleton :
    myfolder = os.path.dirname(paradat.progname)
    if paradat.proglang == 'r':
        progtext = myfolder + os.sep + "template.r"
    else:  ##  assume py
        progtext = myfolder + os.sep + "template.py"
    print("skeleton program to be sought on :",progtext)
    if os.path.isfile(progtext):
        nullprog = getprog(paradat,progtext)
        ##  omits lines that don't belong in paradat.rulemode.
    else:
        raise RuntimeError("Rule template file "+progtext+" not found!")
    ##  for p in nullprog:
        ##  print(p)
    
    ##  showobj(paradat)

    ##  o/p of rule in chosen programming language :
    if paradat.proglang == 'r':
        routlist = rcoding(paradat,mainrule,nullprog)
    else:
        routlist = pycoding(paradat,mainrule,nullprog)
    for r in routlist:
        ##  print(r)
        print(r,file=paradat.progfile)
    print("program code written to file",os.path.abspath(paradat.progfile.name))
    paradat.progfile.close()

    print(NL,file=paradat.listfile)  ##  insert blanks
    parashow(paradat,paradat.listfile)
    paradat.listfile.close()
    ##  parashow(paradat,fo) ; fo.close()
    
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff

    pybyebye()
    return  0




##  PYCODING :
def pycoding (paradat,rule,nullprog):
    "rule conversion routine for python, nullprog is template."

    proglist = [] ; tix2 = "``"
    nullsize = len(nullprog)
    for j in range(12):
        proglist.append(nullprog[j])
        if proglist[-1].startswith("beag_gold"):
            break
    proglist[2] = proglist[2].replace(tix2,rule.id)
    proglist[3] = proglist[3].replace(tix2,rule.dateline)
    ##  amend in herb & leaf to give rule own date!
    dateline = time.asctime(time.localtime(time.time()))
    proglist[4] = proglist[4].replace(tix2,dateline)
    while not nullprog[j].startswith("def beag_stabprep"):
        proglist.append(nullprog[j+1])
        j += 1
    j += 1
    proglist.append(nullprog[j])
    proglist.append('')
    proglist.append(nullprog[j+1])
    stablist = stabcode(rule,paradat.proglang.lower())  ##  convert sigtable
    proglist.extend(stablist)
    j += 1
    while not proglist[-1].endswith("stabprep ends."):
        j += 1
        proglist.append(nullprog[j])
    ##  print('@',j,nullprog[j])
    ##  next comes decision function :
    j += 1
    while not proglist[-1].endswith("ending."):  ##  final comment line.
        proglist.append(nullprog[j])
        ##  deal with amendments :
        if "##  target :" in proglist[-1]:
            proglist[-1] = proglist[-1].replace(tix2,rule.targval)
        elif "rule mode is" in proglist[-1]:
            proglist[-1] = proglist[-1].replace(tix2,rule.rulemode)
        elif proglist[-1].endswith("dump catlist."):
            px = str(rule.catlist)
            proglist[-1] = Htab + "catlist = " + px
        elif proglist[-1].endswith("dump priorvec."):
            px = str(rule.priorvec)
            proglist[-1] = Htab + "priorvec = " + px
        elif proglist[-1].endswith("dump subrules."):
            px = "subrules = " + str(len(rule.testlist))
            proglist[-1] = Htab + px
            proglist.append("\trule = [0] * "+str(len(rule.testlist)))
        elif "compute rule values" in proglist[-1]:
            rv = outtests(paradat,rule.testlist,"vals.")
            proglist.extend(rv)
        j += 1
        if j >= nullsize:
            break  ##  sign of error ?

    return  proglist


##  RCODING :
def rcoding (paradat,rule,nullprog):
    "rule conversion routine for r, nullprog is template."

    proglist = [] ; tix2 = "``"
    nullsize = len(nullprog)
    for j in range(12):
        proglist.append(nullprog[j])
        if proglist[-1].startswith("beag_gold"):
            break
    proglist[2] = proglist[2].replace(tix2,rule.id)
    proglist[3] = proglist[3].replace(tix2,rule.dateline)
    ##  amend in herb & leaf to give rule own date!
    dateline = time.asctime(time.localtime(time.time()))
    proglist[4] = proglist[4].replace(tix2,dateline)
    while not nullprog[j].startswith("beag_stabprep"):
        proglist.append(nullprog[j+1])
        j += 1
    j += 1
    proglist.append(nullprog[j])
    proglist.append('')
    proglist.append(nullprog[j+1])
    stablist = stabcode(rule,paradat.proglang.lower())  ##  convert sigtable
    proglist.extend(stablist)
    j += 1
    while not proglist[-1].endswith("stabprep ends."):
        j += 1
        proglist.append(nullprog[j])
    ##  print('@',j,nullprog[j])
    ##  next comes decision function :
    j += 1
    while not proglist[-1].endswith("ending."):  ##  final comment line.
        proglist.append(nullprog[j])
        ##  deal with amendments :
        if "##  target :" in proglist[-1]:
            proglist[-1] = proglist[-1].replace(tix2,rule.targval)
        elif "rule mode is" in proglist[-1]:
            proglist[-1] = proglist[-1].replace(tix2,rule.rulemode)
        elif proglist[-1].endswith("dump catlist."):
            px = str(rule.catlist)
            px = px.replace('[',"c(")
            px = px.replace(']',')')
            proglist[-1] = Htab + "catlist = " + px
        elif proglist[-1].endswith("dump priorvec."):
            px = str(rule.priorvec)
            px = px.replace('[',"c(")
            px = px.replace(']',')')
            proglist[-1] = Htab + "priorvec = " + px
        elif proglist[-1].endswith("dump subrules."):
            px = "subrules = " + str(len(rule.testlist))
            proglist[-1] = Htab + px
        elif "compute rule values" in proglist[-1]:
            rv = outtests(paradat,rule.testlist)
            proglist.extend(rv)
        j += 1
        if j >= nullsize:
            break  ##  sign of error ?

    return  proglist


##  OUTTESTS :
def outtests (paradat,testlist,vdollar="vals$"):
    "converts a list of rule-tests to list of prog-lines."

    outlist = []
    incr = 0
    if Buck in vdollar:  ##  assume r
        incr = 1
    for j,t in enumerate(testlist):
        lh = "rule[" + str(j+incr) + "] = "
        ##  ts = treelib.tree2str(t,paradat)
        ##  might add as comment, for checking ?
        if incr > 0:  ##  r
            rh = treetor(paradat,t,vdollar)
        else:  ##  assume py
            rh = treetopy(paradat,t,vdollar)
        outlist.append(Htab + lh + rh)

    return  outlist


##  TREETOR :
def treetor (paradat,tree,vdollar="val$"):
    "converts rule tree into r code."

    tout = ''
    if tree.type == "null":
        return  tout
    if tree.type == "numeric":
        return  str(tree.item)
    if tree.type == "feature":
        return  vdollar + paradat.taglist[tree.item]
    if tree.type == "stringy":
        f,s = tree.item  ##  unpack
        f = vdollar + paradat.taglist[f]
        if Quotes2 in s:
            s = Quotes1 + s + Quotes1
        else:
            s = Quotes2 + s + Quotes2
        ##  could still go wrong if s contains both kinds of quote !
        return  "grepl(" + s + ',' + f + Rbra
    ##  must be operation :
    if tree.type == "monadic":
        assert tree.item in treelib.Monadics
        op = tree.item
        v = treetor(paradat,tree.rh,vdollar)
        if op == treelib.Isnt:
            return  Lbra + v + "<= 0" + Rbra
        if op == treelib.Negative:
            return  '-' + v
        ##  check for bracket opening v, to avoid superfluity :
        b1 = Lbra ; b2 = Rbra
        if v.startswith(b1) and v.endswith(b2):
            b1 = '' ; b2 = ''
        if op == "Cosi":
            return  "cos" + b1 + v + b2
        if op == "Fabs":
            return  "abs" + b1 + v + b2
        if op == "Root":
            return  "beag_root" + b1 + v + b2
        if op == "Sine":
            return  "sin" + b1 + v + b2
        if op == "Slog":
            return  "beag_slog" + b1 + v + b2
        if op == "Tanh":
            return  "tanh" + b1 + v + b2
    ##  must be dyadic :
    if tree.type == "bool":
        v1 = treetor(paradat,tree.lh,vdollar)
        v2 = treetor(paradat,tree.rh,vdollar)
        op = tree.item
        if tree.item == treelib.Or:
            return  Lbra + v1 + " | " + v2 + Rbra
        if op == treelib.Both:
            return  Lbra + v1 + " & " + v2 + Rbra
        if op == treelib.Exor:
            return  "beag_exor(" + v1 + ',' + v2 + Rbra
        ##  else error !
        return  Dustbin
    assert tree.type in ["comp","math"]
    v1 = treetor(paradat,tree.lh,vdollar)
    v2 = treetor(paradat,tree.rh,vdollar)
    op = tree.item
    if op == treelib.Maxi:
        return  "max(" + v1 + ',' + v2 + Rbra
    if op == treelib.Mini:
        return  "min(" + v1 + ',' + v2 + Rbra
    if op == treelib.Eq:
        op = "=="
    return  Lbra + v1 + ' ' + op + ' ' + v2 + Rbra
    ##  finally reached operators that work in both languages !


##  TREETOPY :
def treetopy (paradat,tree,vdollar="val."):
    "converts rule tree into python3 code."

    tout = ''
    if tree.type == "null":
        return  tout
    if tree.type == "numeric":
        return  str(tree.item)
    if tree.type == "feature":
        return  vdollar + paradat.taglist[tree.item]
    if tree.type == "stringy":
        f,s = tree.item  ##  unpack
        f = vdollar + paradat.taglist[f]
        if Quotes2 in s:
            s = Quotes1 + s + Quotes1
        else:
            s = Quotes2 + s + Quotes2
        ##  could still go wrong if s contains both kinds of quote !
        return  Lbra + s + " in " + f + Rbra
    ##  must be operation :
    if tree.type == "monadic":
        assert tree.item in treelib.Monadics
        op = tree.item
        v = treetopy(paradat,tree.rh,vdollar)
        if op == treelib.Isnt:
            return  Lbra + v + "<= 0" + Rbra
        if op == treelib.Negative:
            return  '-' + v
        ##  check for bracket opening v, to avoid superfluity :
        b1 = Lbra ; b2 = Rbra
        if v.startswith(b1) and v.endswith(b2):
            b1 = '' ; b2 = ''  ##  avoid superfluous parentheses
        if op == "Cosi":
            return  "math.cos" + b1 + v + b2
        if op == "Fabs":
            return  "abs" + b1 + v + b2
        if op == "Root":
            return  "beag_root" + b1 + v + b2
        if op == "Sine":
            return  "math.sin" + b1 + v + b2
        if op == "Slog":
            return  "beag_slog" + b1 + v + b2
        if op == "Tanh":
            return  "math.tanh" + b1 + v + b2
    ##  must be dyadic :
    if tree.type == "bool":
        v1 = treetopy(paradat,tree.lh,vdollar)
        v2 = treetopy(paradat,tree.rh,vdollar)
        op = tree.item
        if op == treelib.Exor:
            return  "beag_exor(" + v1 + ',' + v2 + Rbra
        if v1.startswith(Lbra) and v1.endswith(Rbra):
            v1 = "beag_bool" + v1
        else:
            v1 = "beag_bool(" + v1 + Rbra
        if v2.startswith(Lbra) and v2.endswith(Rbra):
            v2 = "beag_bool" + v2
        else:
            v2 = "beag_bool(" + v2 + Rbra
        if tree.item == treelib.Or:
            return  Lbra + v1 + " or " + v2 + Rbra
        if op == treelib.Both:
            return  Lbra + v1 + " and " + v2 + Rbra

        ##  else error !
        return  Dustbin
    if not (tree.type in ["comp","math"]):
        ts = treelib.tree2str(tree,paradat)
        print(ts)
        raise RuntimeError("unexpected type "+tree.type+" in treetopy() !!")
    v1 = treetopy(paradat,tree.lh,vdollar)
    v2 = treetopy(paradat,tree.rh,vdollar)
    op = tree.item
    if op == treelib.Maxi:
        return  "max(" + v1 + ',' + v2 + Rbra
    if op == treelib.Mini:
        return  "min(" + v1 + ',' + v2 + Rbra
    if op == treelib.Eq:
        op = "=="
    return  Lbra + v1 + ' ' + op + ' ' + v2 + Rbra
    ##  finally reached operators that work in both languages !


##  STABCODE :
def stabcode (rule,proglang):
    "converts sigtable to lines in programming language."

    stablist = []
    sr = len(rule.stab)
    sw = len(rule.catlist)
    nullcell = [0] * sw
    stabkeys = sorted(rule.stab.keys())
    ##  print(stabkeys)
    for j,sk in enumerate(stabkeys):
        ##  print(sk,int(sk,2),rule.stab[sk])
        ro = rule.stab[sk]
        if sum(ro) <= 0:  ##  null cells cause trouble in r
            ro = nullcell
        lout = str(ro)
        lout = "stab[[" + repr(sk) + "]] = ~" + lout
        if proglang == 'r':
            lout = lout.replace("~[","c(")
            lout = lout[:-1] + ')'
        else:  ##  assume py
            lout = lout.replace("~[",'[')
            lout = lout.replace("[[",'[')
            lout = lout.replace("]]",']')
        stablist.append(Htab+lout)
        
    return  stablist


##  GETPROG :
def getprog (paradat,progtext):
    "gets program skeleton from file named by progtext."

    fi = open(progtext,'r',encoding="utf-8")
    rawlines = fi.readlines()
    fi.close()
    ##  more convenient without trailing newlines :
    datlines = []
    for r in rawlines:
        if paradat.rulemode == "demonic" and "omit if demonic" in r:
            continue
        if paradat.rulemode == "tabular" and "omit if tabular" in r:
            continue  ##  skip inappropriate instructions.
        datlines.append(r.rstrip())  ##  leaves front alone

    return  datlines


##  MODFRAX :
def modfrax (freqvec,digs=7):
    "attenuates frequencies to produce bayesian multipliers."

    n = sum(freqvec) ; c = len(freqvec)
    slug = Gold
    pvec = [f+slug/c for f in freqvec]
    psum = sum(pvec)

    return  [round(p/psum,digs) for p in pvec]




##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
