"Python3 dataset splitter, to precede herb.py."

##  SEED :
##
##  Simple dataset splitter in Python3 (Selectively Extracts Example Data).
##  Copyright (c), 2016, R.S. Forsyth.
##  ancestral progenitor, June 1985.
##  first version : 04/12/2015
##  last revision : 08/08/2016


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  copy
import  math
import  os, sys
import  random, time

##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  parashow
from glib import  showobj
from glib import  openfile
from glib import  pybyebye
from glib import  request
from gridlib import  cols2use, matread
from gridlib import  matdump, vecline
from gridlib import  var2col
from gridlib import  bastatz, midvalue
##  from gridlib import  vecdist
from gridlib import  copygrid, distest
from gridlib import  matshow, precrec
from gridlib import  sckappa2
from gridlib import  hedgesg


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ##  ("atomize",1,0,1),
    ("randseed",0,0,999999999),  ##  random seed, <2 means haphazard seed
    ("datfrac",1.0/Gold,0,1),  ##  proportion to training sample
    ("jobname","beagle16"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ("maindat",''),  ##  file of input data
    ##  ("cols2use",''),  ##  variables to be used
    ("targval",''),  ##  target variable
    ("targcol",0),  ##  target column number
    ##  ("idfield",''),  ##  varname for o/p label
    ("traindat",''),  ##  first output file (training)
    ("testdat",''),  ##  second output file (holdout)
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"seed.txt"),  ##  logfile ??
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "Simple Exploratory Example Distributor, v1.3"
##  Gold = 5.0**0.5*0.5+0.5
Mincases = 4
Dattail = "_dat1"
Testtail = "_dat2"
Popsize = 233
Dustbin = "??!!"
##  Outpars = ["dateline","progname","id","testfile","targvar"]


##  Main program :
##
def main():
    "Main line."
    
    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time() ; clockon = time.clock()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()
    if paradat.randseed >= 2:  ##  initialize random-number generator early
        random.seed(paradat.randseed)
    else:
        random.seed(None)  ##  haphazard

    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if not paradat.outpath.endswith(os.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
        if not os.path.isdir(paradat.outpath):  ##  try another outpath
            paradat.outpath = os.getcwd() + os.sep + "op" + os.sep
            if not os.path.isdir(paradat.outpath):
                os.makedirs(paradat.outpath)  ##  create o/p folder
            paradat.outfile = paradat.outpath + os.path.basename(paradat.outfile)

    ##  prepare input file :
    if not os.path.isdir(paradat.outpath):
        os.makedirs(paradat.outpath)
        print(paradat.outpath,"created.")
    if paradat.maindat and os.path.isfile(paradat.maindat):
        print("data to be read from",paradat.maindat)
    else:
        raise RuntimeError("No such file as "+paradat.maindat)
    ##  won't be opened till getfile() .
    ##  prepare output files :
    if paradat.traindat:
        training = paradat.traindat
    else:
        training = paradat.outpath + paradat.jobname + Dattail + ".dat"
    paradat.traindat = open(training,'w',encoding="utf8")
    if paradat.testdat:
        tf = paradat.testdat
    else:
        tf = paradat.outpath + paradat.jobname + Testtail + ".dat"
    paradat.testdat = open(tf,'w',encoding="utf-8")
    ##  already opened.

    ##  read input data :
    ##  vars 2b skipped will affect paradat.goodcols :
    paradat.skipvars = ''  ##  to avoid column omission
    r = getfile(paradat,paradat.maindat)
    if r < 0:
        raise RuntimeError("Problem reading data file : "+paradat.maindat+'!')

    ##  parashow(paradat)

    ##  prepare o/p :
    headline = Htab.join(paradat.taglist)
    print(headline,file=paradat.traindat)
    print(headline,file=paradat.testdat)
    rownums = list(range(paradat.rows))
    random.shuffle(rownums)
    ##  print(rownums)
    flipcase = paradat.datfrac * paradat.rows
    ix = round(flipcase)
    ##  print(flipcase,ix)
    set1 = set(rownums[:ix])
    set2 = set(rownums[ix:])
    ##  print(len(set1),len(set2))
    ##  print(sorted(set1))
    ##  print(sorted(set2))
    n1 = n2 = 0
    ##  main loop :
    for i,j in enumerate(rownums):
        ##  rowvals = paradat.datamat[j]  ##  doesn't preserve input sequence
        rowvals = paradat.datamat[i]
        lout = vecline(rowvals,7)
        ##  if i <= flipcase:
        if i in set1:  ##  preserve input sequence
            fo = paradat.traindat
            n1 += 1
        else:
            fo = paradat.testdat
            n2 += 1
        print(lout,file=fo)    

    ##  o/p of results :
    print("data rows processed =",paradat.rows)
    pout = open(paradat.outfile,'w',encoding="utf-8")
    parashow(paradat,pout)
    pout.close() ; nc = n1
    for f in [paradat.traindat,paradat.testdat]:
        fn = os.path.abspath(f.name)
        print(nc,"cases sent to",fn)
        nc = n2
        f.close()
    
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff
    pybyebye()
    return  0




##  GETFILE :
def getfile (paradat,fi):
    "obtains data grid from file."

    paradat.datamat,paradat.taglist,paradat.tagdict = matread(fi,paradat)
    paradat.rows = len(paradat.datamat)
    print("data rows =",paradat.rows)
    paradat.cols = len(paradat.taglist)
    print("data cols =",paradat.cols)
    print("column names :")
    print(paradat.taglist)
    if paradat.rows < Mincases:  ##  not enough data
        print("Not enough data cases !")
        print("Program halts.")
        return  -1
    goodcols = cols2use(paradat)
    paradat.goodcols = goodcols
    paradat.goodlist = sorted(paradat.goodcols)
    if len(goodcols) < 1:
        print("No numeric variables found in ",paradat.datfile,'!')
        print("Program halts.")
        return  -2

    return  0  ##  success, paradat altered




##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
