"Python3 tree-tidying stuff for BEAGLE & RUNSTER."

##  TIDYLIB :
##
##  BEAGLE tree-tidying software (in Python3) :
##  Copyright 2016, R.S. Forsyth.
##  ancestral progenitor : Sept, 1980.
##  first version : 15/08/2015
##  last revision : 12/09/2016


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


##  pythonic libraries
import  math
import  os, sys
import  random, time
import  copy


##  local libraries :
from globdefs import *
##  Sack() 2b used as base type.
from glib import  numeric, numerize
from glib import  showobj

import  treelib
newnode = treelib.newnode
##  other functions 2b be prefixed.


Maxdigs = 15  ##  could be smaller (bool/comp only)  ??
Bset = set(["bool","comp"])  ##  operators that yield booleans


##  SAMEVAL :
def sameval (a1,a2):
    "tests whether 2 atomic nodes are equal."

    assert not (a1.lh or a1.rh)
    assert not (a2.lh or a2.rh)
    if a1.type != a2.type:
        return  False
    ## else:
    return  a1.item == a2.item


##  SAMETREE :
def sametree (t1,t2):
    "tests whether 2 trees are equivalent."

    if not t1:
        return  not t2  ##  if null link passed in
    if not t2:
        return  not t1
    if not (t1.lh or t1.rh) and not (t2.lh or t2.rh):  ##  2 atoms
        return  sameval(t1,t2)
    assert  t1.rh or t2.rh
    ##  gets here if at least 1 is composite :
    if t1.item == t2.item:  ##  same op
        assert t1.type == t2.type
        s1 = sametree(t1.lh,t2.lh) and sametree(t1.rh,t2.rh)
        if s1:
            return  s1
        ##  also consider associative operators :
        if t1.item in treelib.Balanced:
            return  sametree(t1.lh,t2.rh) and sametree(t1.rh,t2.lh)

    return  False


##  TREETIDY :
def treetidy (paradat,t):
    "simplifies tree t by expressing it more succinctly."

    if not t:
        return  t  ##  possible ??
    atom = (not t.lh) and (not t.rh)
    if atom:
        return  treelib.copytree(t)
    elif  t.item in treelib.Monadics:
    ##  gets here if composite :
        v = treetidy(paradat,t.rh)
        return  montidy(paradat,newnode('',t.item,v,t.type))
    elif  t.item in treelib.Dyadics:  ##  dyadic op :
        v1 = treetidy(paradat,t.lh)
        v2 = treetidy(paradat,t.rh)
        dx = newnode(v1,t.item,v2,t.type)
        if dx.item in treelib.Boolops:
            return  booltidy(paradat,dx)
        elif dx.item in treelib.Compops:
            return  comptidy(paradat,dx)
        elif dx.item in treelib.Mathops:
            return  mathtidy(paradat,dx)
        else:
            return  dx  ##  dummy at present

    return  treelib.copytree(t)  ##  if uncaught, just copy


##  BOOLTIDY :
def booltidy (paradat,t):
    "simplifies boolean node/tree t."

    assert  t.type == "bool"
    ##  .lh & .rh should already have been tidied !
    if t.lh.type == "numeric" and t.rh.type == "numeric":
        v = treelib.boolval(t.lh.item,t.item,t.rh.item)
        return  newnode('',round(v,Maxdigs),'',"numeric")
    ##  gets here if at most 1 operand is elementary :
    if t.item == treelib.Both:  ##  logical and
        if t.lh.type == "numeric" and t.lh.item <= 0.0:
            return  newnode('',0,'',"numeric")
        if t.lh.type == "numeric" and t.lh.item > 0 and t.rh.type in Bset:
            return  t.rh
        if t.rh.type == "numeric" and t.rh.item <= 0.0:
            return  newnode('',0,'',"numeric")
        if t.rh.type == "numeric" and t.rh.item > 0 and t.lh.type in Bset:
            return  t.lh
        if sametree(t.lh,t.rh):
            return  newnode(t.rh,treelib.Gt,newnode('',0,'',"numeric"),"comp")
    elif t.item == treelib.Or:  ##  logical or
        if t.lh.type == "numeric" and t.lh.item <= 0.0 and t.rh.type in Bset:
            return  t.rh
        if t.rh.type == "numeric" and t.rh.item <= 0.0 and t.lh.type in Bset:
            return  t.lh
        if t.lh.type == "numeric" and t.lh.item > 0:
            return  newnode('',1,'',"numeric")
        if t.rh.type == "numeric" and t.rh.item > 0:
            return  newnode('',1,'',"numeric")
        if sametree(t.lh,t.rh):
            return  newnode(t.lh,treelib.Gt,newnode('',0,'',"numeric"),"comp")
    elif t.item == treelib.Exor:
        if sametree(t.lh,t.rh):
            return  newnode('',0,'',"numeric")
        elif t.lh.type == "numeric" and t.lh.item <= 0:
            return  newnode(t.rh,treelib.Gt,newnode('',0,'',"numeric"),"comp")
        elif t.rh.type == "numeric" and t.rh.item <= 0:
            return  newnode(t.lh,treelib.Gt,newnode('',0,'',"numeric"),"comp")
        elif t.lh.type == "numeric" and t.lh.item > 0:
            return  newnode('',treelib.Isnt,t.rh,"monadic")
        elif t.rh.type == "numeric" and t.rh.item > 0:
            return  newnode('',treelib.Isnt,t.lh,"monadic")
        else:
            return  t
        
    return  t  ##  fallback


##  COMPTIDY :
def comptidy (paradat,t):
    "simplifies comparison node/tree t."

    assert  t.type == "comp"
    ##  .lh & .rh should already have been tidied !
    if t.lh.type == "numeric" and t.rh.type == "numeric":
        v = treelib.compval(t.lh.item,t.item,t.rh.item)
        return  newnode('',round(v,Maxdigs),'',"numeric")
    if t.item == treelib.Eq:
        if sametree(t.lh,t.rh):
            return  newnode('',1,'',"numeric")
        else:
            return  t
    if t.item in [treelib.Gt,treelib.Lt]:
        if sametree(t.lh,t.rh):
            return  newnode('',0,'',"numeric")
        else:
            return  t

    return  t  ##  fallback (t is already a fresh node)


##  MATHTIDY :
def mathtidy (paradat,t):
    "simplifies mathematic operation t."

    assert  t.type == "math"
    ##  .lh & .rh should already have been tidied !
    if t.lh.type == "numeric" and t.rh.type == "numeric":
        v = treelib.mathval(t.lh.item,t.item,t.rh.item)
        return  newnode('',v,'',"numeric")
    ##  gets here if at most 1 operand is elementary :
    if t.item == treelib.Mult:  ##  multiplication
        if t.lh.type == "numeric" and t.lh.item == 0.0:
            return  newnode('',0,'',"numeric")
        if t.lh.type == "numeric" and t.lh.item == 1:
            return  t.rh
        if t.rh.type == "numeric" and t.rh.item == 0.0:
            return  newnode('',0,'',"numeric")
        if t.rh.type == "numeric" and t.rh.item == 1:
            return  t.lh
    if t.item == treelib.Plus:
        if t.lh.type == "numeric" and t.lh.item == 0.0:
            return  t.rh
        if t.rh.type == "numeric" and t.rh.item == 0:
            return  t.lh
        ##  change + - or + ~ to - ??
    if t.item == treelib.Less:
        if t.rh.type == "numeric" and t.rh.item == 0:
            return  t.lh
        ##  change - - or - ~ to + ??
    if t.item in [treelib.Mini,treelib.Maxi]:
        if sametree(t.lh,t.rh):
            return  t.lh
        else:
            return  t
        
    return  t  ##  fallback (t is already a fresh node)


##  MONTIDY0 :
def montidy0 (paradat,t):
    "tidy tree t based on monadic operator."

    assert t.type == "monadic"
    ##  .rh assumed 2b tidied in caller :
    if t.item == treelib.Isnt:  ##  logical negation
        if t.rh.type == "numeric":
            if t.rh.item > 0:
                return  newnode('',0,'',v.type)
            else:
                return  newnode('',1,'',"numeric")
        else:
            return  t  ##  should be result of treetidy
    elif t.item == treelib.Negative:  ##  arithmetic minus
        if t.rh.item == treelib.Negative:  ##  double minus
            return  t.rh.rh  ##  must exist (if well-formed!)
    elif t.rh.type == "numeric":  ##  operand is atomic
        ##  evaluate operation :
        v = treelib.leibniz(t.item,t.rh.item)  ##  might hav2 keep many digits
        return  newnode('',v,'',"numeric")
    else:  ##  can't be simplified
        return  t  ##  no need to copy (since already newnode) ?

    ##  can't fall thru.


##  MONTIDY :
def montidy (paradat,t):
    "tidy tree t based on monadic operator."

    assert t.type == "monadic"
    ##  .rh assumed 2b tidied in caller.
    if t.rh.type == "numeric":  ##  operand is atomic
        ##  evaluate operation :
        v = treelib.leibniz(t.item,t.rh.item)  ##  might hav2 keep many digits
        return  newnode('',v,'',"numeric")
    ##  .rh isn't atomic :
    if t.item == "Fabs" and t.rh.item == "Fabs":
        return  t.rh.rh  ##  must exist (if well-formed!)
    if t.item == treelib.Isnt:  ##  logical negation
        if t.rh.type == "numeric":  ##  already covered ?
            if t.rh.item > 0:
                return  newnode('',0,'',"numeric")
            else:
                return  newnode('',1,'',"numeric")
        else:
            return  t  ##  should be result of treetidy
    if t.item == treelib.Negative and t.rh.item == treelib.Negative:
        return  t.rh.rh  ##  must exist (!?)
    return  t  ##  no obvious simplification .




Varnames = ["zero","unit","pair","four","luck","Minus99","nine",
            "Zero","sub1","half","text","name"]


##  test blocks :

if 0 > 0:  ##  testing
    print(Varnames)
    ##  pseudo-context :
    paradat = Sack()
    paradat.varnames = Varnames
    paradat.taglist = Varnames
    paradat.thisrow = [0,1,2,4,7,-99,9,0,-1,0.5,"Hello World!","Pythonic"]
    paradat.cols = len(paradat.varnames)
    assert len(paradat.thisrow) == paradat.cols
    paradat.allcols = list(range(paradat.cols))
    paradat.skipcols = set([0,8])  ##  forbidden
    paradat.goodcols = set(paradat.allcols) - paradat.skipcols
    paradat.strcols = [10,11]
    paradat.goodcols = paradat.goodcols - set(paradat.strcols)
    paradat.goodlist = sorted(paradat.goodcols)  ##  to save repeated re-listing
    ##  specimen string constants :
    paradat.strdict = {10: ["ello",'!',"zonk","lo","Hell","wo",'W'],
                       11: ["ic",'P','y',"zonq","xxzz",'@']}
    paradat.cols2use = sorted(paradat.goodlist + paradat.strcols)
    print(paradat.goodcols,paradat.goodlist,paradat.cols2use)


if 0 > 0:
    ##  basic test of tidying :
    ##  ts = "(( pair * 1 ) & ($Root 66 & four))"
    ##  ts = "($Root (2 + luck) | $Root (luck + 2))"
    ##  ts = "(pair + ( pair ^ pair )) "
    ##  ts = "(( nine < 18.3590286) * (3 & 3.904023))"
    ##  ts = "( ( nine < 18.36 ) * 1 )"
    ##  ts = "( 0 + ( nine < 18.36 ) )"
    ##  ts = "( four > ( nine > ( $Sine four & -235 )))"
    ##  ts = "( ( ( nine < 12.0608627 ) + $Tanh -15.2050097 ) + ( nine < 14.2063447 ) )"
    ##  ts = "( ( $! (449 - luck) - (nine < nine) ) )"
    ##  ts = "( ($Fabs $Fabs 12.75 ) + $~ $~ (nine > nine))"
    ##  ts = "( $Fabs $Fabs (nine = nine)  + $~ $~ $Slog(unit \ unit) )"
    ##  ts = "( 568 & (pair > 0))"
    ##  ts = "( ( luck > 0) & $Sine 1.44)"
    ##  ts = "$Cosi 1 + $Sine 1"
    ##  ts = "( unit & ( luck < ( luck < $~ -16.9157046 ) ) )"
    ##  ts = "$! ( -271 ; (luck > half ) & (unit ; -99 ) )"
    ts = "( ( 19.48 & (nine > four) ) = unit)"
    tx = treelib.str2tree(paradat,ts)
    for j in range(4):
        if j == 0:
            t1 = tx
        else:
            t1 = treelib.maketree(paradat,6)
        v1 = treelib.treeval(t1,paradat,paradat.thisrow)
        s1 = treelib.tree2str(t1,paradat)
        print(s1)
        t2 = treetidy(paradat,t1)
        v2 = treelib.treeval(t2,paradat,paradat.thisrow)
        s2 = treelib.tree2str(t2,paradat)
        print(s2)
        print(v1,v2)
        if abs(v1-v2) > 1e-10:
            print("!!!!")
            ##  break
        paradat.thisrow[4] = random.randint(-99,100)
