"Python3 tree-processing stuff for BEAGLE."

##  TREELIB :
##
##  BEAGLE tree-handling software (in Python3) :
##  Copyright 2016, R.S. Forsyth.
##  ancestral progenitor : Sept, 1980.
##  first version : 15/08/2015
##  last revision : 02/08/2016


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  math
import  os, sys
import  random, time
import  copy


##  local libraries :
from globdefs import *
##  Sack() 2b used as base type.
from glib import  numeric, numerize
from glib import  showobj


##  globals :
Buck = '$'
Lt = '<'
Gt = '>'
Eq = '='
Exclaim = '!'
Negative = '~'
Isnt = Exclaim
Vertical = '|'
Any1 = Vertical
Or = Any1
Plus = '+'
Dash = '-'
Less = Dash
Maxi = '^'
Mini = '\\'
At = '@'
Also = '&'
Both = Also
Exor = ';'  ##  Exor
Lbra = '('
Rbra = ')'
Star = '*'
Mult = Star
Backtick = '`'
Contains = '?'
##  Open/Shut: [] ??
Null = "null"
Dollars = "$$"
Typelist = ["null","bool","comp","feature","math","numeric","stringy"]
Mathops = [Plus,Less,Mult,Maxi,Mini]
Compops = [Eq,Gt,Lt]
Boolops = [Or,Both,Exor]
Monadics = [Isnt,Negative,"Cosi","Fabs","Root","Sine","Slog","Tanh"]
Dyadics = Boolops + Compops + Mathops
Opchars = Dyadics + Monadics  ##  -> Opcodes
Balanced = Boolops + [Eq] + [Maxi,Mini,Mult]
Optable = {"bool": Boolops,"comp": Compops,"math": Mathops}

##  support functions :


##  OPCLASS :
def opclass (op):
    "returns the kind of an operator."

    if op in Monadics:
        return  "monadic"
    if op in Boolops:
        return  "bool"
    if op in Compops:
        return  "comp"
    if op in Mathops:
        return  "math"
    if op == Contains:
        return  "stringy"

    return  "null"  ##  "noop" ?


##  BOOLVAL :
def boolval (v1,op,v2):
    "evaluates boolean operation."

    v = 0
    if v2 == '':
        return  0  ##  error !
    if op == Or:
        v = (v1 > 0.0) or (v2 > 0.0)
    elif op == Both:
        v = (v1 > 0.0) and (v2 > 0.0)
    elif op == Exor:
        v = (v1 > 0.0) != (v2 > 0.0)
    else:  ##  Negative ?? (Nand?)
        v = (v2 <= 0.0)

    return  v + 0  ##  makes boolean binary (0,1)


##  COMPVAL :
def compval (v1,op,v2):
    "evaluates comparison operation."

    if v2 == '':  ##  error ?!
        return  0
    v = 0
    if op == Eq:
        v = (v1 == v2)
    elif op == Gt:
        v = (v1 > v2)
    else:  ##  Lt
        v = (v1 < v2)

    return  v + 0  ##  coerce to numeric


##  MATHVAL :
def mathval (v1,op,v2):
    "evaluates arithmetic operation."

    if v2 =='':  return  0.0  ##  error ?!
    v = 0.0
    if op == Plus:
        v = v1 + v2
    elif op == Less:
        v = v1 - v2
    elif op == Mult:
        v = v1 * v2
    elif op == Maxi:
        v = max(v1,v2)
    elif op == Mini:
        v = min(v1,v2)
    ##  else ??

    return  v


##  LEIBNIZ :
def leibniz (op,v):
    "evaluates a monadic operation."
    
    if op == '':  ##  better as single blank ?
        return  v  ##  null op, normal case
    elif op == Isnt:  ##  boolean negation
        return  (v <= 0.0) + 0
    elif op == "Fabs":
        return  abs(v)
    elif op == Negative:  ##  unary minus
        return  -v
    elif op == "Root":
        if v >= 0.0:
            return  v ** 0.5
        else:
            return -math.sqrt(-v)
    elif op == "Slog":
        if v < 0.0:
            return  -math.log1p(abs(v))
        else:
            return  math.log1p(v)
    elif op == "Tanh":
        return  math.tanh(v)
    elif op == "Cosi":
        return  math.cos(v)
    elif op == "Sine":
        return  math.sin(v)

    return  v  ##  fall-back


##  SIZE :
def size (t):
    "computes size of node/tree."

    if not t:
        return  0
    s = 0
    if t.type == "stringy":
        return  2  ##  special case, no subtrees (3?)
    if t.type == "monadic":
        pass
    elif t.lh:
        s += size(t.lh)
    s += 1  ##  content
    if t.rh:
        s += size(t.rh)
        
    return  s

    
##  COPYTREE :
def copytree (t):
    "makes fresh copy of tree t."

    return  copy.deepcopy(t)
    ##  seems 2b adequate.


##  READTREE ??


##  MAKEITEM :
def makeitem (paradat):
    "makes random atomic node; uses paradat.cols2use, .strcols, .strdict"

    node = Sack()
    node.id = id(node)
    node.lh = '' ; node.rh = ''
    r = random.random()
    if r > 0.5:  ##  feature
        r = random.choice(paradat.cols2use)
        ##  quicker eval with index not string.
        nt = "feature"
        if r in paradat.strcols:  ##  string inclusion, special case
            s = random.choice(paradat.strdict[r])
            node.item = (r,s)  ##  feature,string held as pair
            node.type = "stringy"
            return  node
    else:  ##  variable
        r = random.random()
        if r > 0.75:  ## float
            r = (random.random() - random.random()) * 100.0
            r = round(r,7)  ##  avoid spurious precision ?
        elif r > 0.5:  ##  digit
            r = random.randint(0,9)
        elif r > 0.25:  ##  integer
            r = random.randint(-256,256)
        else:  ##  larger integer
            r = random.randint(-1024,1024)
        nt = "numeric"
    node.item = r
    node.type = nt

    return  node


##  NEWNODE :
def newnode (head,core,tail,kind):
    "constructs new node from given values; quasi __init__."

    node = Sack()
    node.id = id(node)  ##  ever used ??
    node.lh = head
    node.item = core
    node.rh = tail
    node.type = kind

    return  node


##  SHOWITEM :
def showitem (node,paradat,fo=sys.stdout):
    "basic tree/node output."

    content = node.item
    if node.type == "feature":
        content = paradat.taglist[content]  ##  feature name
    print(content,file=fo)
    print(node.type,file=fo)
    print(node.lh,node.rh,file=fo)
    print(Dollars,file=fo)

    return


##  MAKETREE :
def maketree (paradat,dmax=8):
    "makes an expression tree, maxdepth=dmax."

    r = random.randint(0,10)
    if dmax < 1 or r < 2:
        return  makeitem(paradat)
    if r < 4:
        return  makemath(paradat,dmax-1)
    elif r < 6:
        return  makebool(paradat,dmax-1)
    elif r < 8:
        return  makecomp(paradat,dmax-1)
    else:  ##  unary
        return  makemono(paradat,dmax-1)

    return  ''  ##  crash!


##  MAKEMONO :
def makemono (paradat,dmax=4):
    "makes a monadic operation."

    node = Sack(time.clock())
    node.item = random.choice(Monadics)
    node.lh = ''
    if dmax < 1 or random.random() < 0.5:
        node.rh = makeitem(paradat)
    else:
        node.rh = maketree(paradat,dmax-1)
    node.type = "monadic"

    return  node

            
##  MAKEMATH :
def makemath (paradat,dmax=7):
    "makes a math term."

    if dmax < 1:
        return  makeitem(paradat)
    r = random.randint(0,9)
    if r < 4:
        return  makeitem(paradat)
    node = Sack(time.time())
    ##  subexpression :
    rx = random.randint(1,2)
    if rx == 1:
        node.lh = makeitem(paradat)
    else:
        node.lh = makemath(paradat,dmax-1)
    node.item = random.choice(Mathops)
    node.type = "math"
    r = random.randint(0,9)
    if r < 4:
        node.rh = makeitem(paradat)
    elif r < 7:
        node.rh = makemath(paradat,dmax-1)
    else:
        node.rh = maketree(paradat,dmax-1)

    return  node


##  MAKECOMP :
def makecomp (paradat,dmax=7):
    "makes a comparison operation."

    if dmax < 1:
        return  makeitem(paradat)
    ##  expression :
    r = random.randint(0,9)
    node = Sack(time.time())
    if r < 2:
        node.lh = makeitem(paradat)
    else:
        node.lh = makemath(paradat,dmax-1)
    node.item = random.choice(Compops)
    r = random.randint(0,9)
    if r < 4:
        node.rh = makeitem(paradat)
    else:
        node.rh = makemath(paradat,dmax-1)
    node.type = "comp"

    return  node


##  MAKEBOOL :
def makebool (paradat,dmax=7):
    "makes a boolean operation."

    if dmax < 1:
        return  makeitem(paradat)
    ##  expression :
    node = Sack(time.time())
    r = random.randint(0,9)
    if r < 2:
        node.lh = makeitem(paradat)
    else:
        node.lh = makemath(paradat,dmax-1)
    node.item = random.choice(Boolops)
    r = random.randint(0,9)
    if r < 4:
        node.rh = makeitem(paradat)
    elif r < 8:
        node.rh = makecomp(paradat,dmax-1)
    else:
        node.rh = makemath(paradat,dmax-1)
    node.type = "bool"

    return  node

    
##  TREE2STR :
def tree2str (t,paradat):
    "makes string form of node t."

    if t.type == "stringy":
        v,s = t.item  ##  split pair
        content = paradat.taglist[v]
        lout = Lbra + ' ' + content + ' ' + Contains + ' '
        lout += Backtick + str(s) + Backtick + Rbra
        return  lout
    atom = (not t.lh) and (not t.rh)
    if atom:  ##  normal atom
        content = t.item
        if t.type == "feature":
            content = paradat.taglist[content]
        return  str(content)
    ##  composite expression :
    if t.type == "monadic":
        lout = Buck + t.item + ' '
        try:
            v = tree2str(t.rh,paradat)
        except AttributeError as xxxx:
            showobj(t) ; print(t.rh)
            print(xxxx)
        return  lout + v
    ##  dyadics :
    lout = Lbra + ' '
    ##  left subtree :
    try:
        left = tree2str(t.lh,paradat)
    except AttributeError as xxxx:
        print('!',t.lh,Backtick)
        showitem(t,paradat)
        print(xxxx)
    lout += str(left) + ' '
    lout += str(t.item) + ' '  ##  core
    lout += str(tree2str(t.rh,paradat)) + ' ' + Rbra
    ##  right subtree & closing bracket.

    return  lout


##  STR2TREE :
def str2tree (paradat,ts):
    "turns fully bracketed tree-string into tree."

    toks = treetox(ts)
    pole = tox2pole(toks)
    tout = poletree(paradat,pole)

    return  tout


##  TREEVAL :
def treeval (tree,paradat,thisrow):
    "evaluates rule tree on current data case, thisrow."

    v = 0.0
    if not tree:
        return  v  ##  impossible ??
    t = tree.type
    x = tree.item
    if t == "null":  return  v  ##  possible ??
    if t == "feature":
        return  thisrow[x]  ##  x is feature index
    if t == "numeric":
        return  x
    if t == "stringy":
        f,s = x  ##  unpack varnum & substring
        v = thisrow[f]
        return  (s in v) + 0
    if t == "monadic":
        v = treeval(tree.rh,paradat,thisrow)
        return  leibniz(x,v)
    ##  composite :
    v1 = treeval(tree.lh,paradat,thisrow)
    v2 = treeval(tree.rh,paradat,thisrow)
    if t == "comp":
        v = compval(v1,x,v2)
    elif t == "math":
        v = mathval(v1,x,v2)
    elif t == "bool":
        v = boolval(v1,x,v2)
    ##  no else ?!

    return  v


##  GETNODES :
def getnodes (tree,paradat,pure=True):
    "unpacks nodes of tree in natural tree order (depth1st?)."

    if not tree:
        return  []
    nodelist = []
    if tree.lh:
        nodelist.extend(getnodes(tree.lh,paradat,pure))
    if pure:
        nodelist.append(tree)  ##  just address pointer
    else:
        nodelist.append(tree2str(tree,paradat))  ##  string version
    if tree.rh:
        nodelist.extend(getnodes(tree.rh,paradat,pure))

    return  nodelist


##  MATE :
def mate (paradat,maternal,paternal):
    "initial crossover routine."

    m = getnodes(maternal,paradat)
    p = getnodes(paternal,paradat)
    opcodes = []
    if random.random() > (Gold-1.0):
        opcodes = [random.choice(Dyadics)]  ##  bias towards dyadics
    if maternal.item in Opchars:
        opcodes.append(maternal.item)
    if paternal.item in Opchars:
        opcodes.append(paternal.item)
    opcodes.append(random.choice(Dyadics))
    if paternal.lh and maternal.lh:
        pass  ##  neither parent monadic
    else:
        opcodes.append(random.choice(Monadics))
    ##  print(opcodes)
    ##  make offspring :
    mx = [id(x) for x in m]
    px = [id(x) for x in p]
    ##  print(len(mx)) ; print(len(px))
    head = random.choice(m)
    tail = random.choice(p)
    node = Sack()
    node.item = random.choice(opcodes)
    if node.item in Monadics:
        node.lh = ''
        node.type = "monadic"
        if random.random() > 0.5:
            node.rh = copytree(head)
        else:
            node.rh = copytree(tail)
    else:  ##  dyadic op
        node.lh = copytree(head)
        node.rh = copytree(tail)
        if node.item in Boolops:
            node.type = "bool"
        elif node.item in Compops:
            node.type = "comp"
        elif node.item in Mathops:
            node.type = "math"
        else:  ##  error !
            node.type = "null"
            
    return  node
        

##  ZONK :
def zonk (paradat,t):
    "initial mutation routine."

    ##  obtain elements of tree t :
    nodelist = getnodes(t,paradat)
    nn = len(nodelist)
    ##  print(nn)
    nz = nodelist[random.randint(0,nn-1)]
    tz = copyzap(paradat,t,nz)
    ##  decide whether to repeat :
    if nn + size(tz) > 8 + random.randint(2,7):
        nodelist = getnodes(tz,paradat)
        nn = len(nodelist)
        nz = nodelist[random.randint(0,nn-1)]
        tz = copyzap(paradat,tz,nz)
                
    return  tz


##  ZAPNODE :
def zapnode (paradat,t):
    "alters individual node of a tree."

    tx = Sack()
    if t.lh:
        tx.lh = copytree(t.lh)
    else:
        tx.lh = t.lh
    if t.rh:
        tx.rh = copytree(t.rh)
    else:
        tx.rh = t.rh
    tx.type = t.type  ##  usually stays same
    if t.item in Dyadics and random.randint(1,20) in [2,19]:
        tx.lh,tx.rh = tx.rh,tx.lh  ##  swap subtrees
        tx.item = t.item
        ##  same type
        return  tx
    if t.type == "feature":
        r = random.random()
        if r < 0.75:  ##  reselect feature
            i = -1
            while not (i in paradat.cols2use):
                i = random.choice(paradat.goodlist)
            tx.item = i
        elif r < 0.89:
            ##  could look into .cuts eventually
            tx = makeitem(paradat)
        else:
            tx = maketree(paradat,4+random.randint(0,4))
    elif t.type == "numeric":
        r = random.random()
        if r > 0.25:
            v = t.item
            rr = random.randint(1,5)
            if rr == 1:
                tx.item = v * 0.96
            elif rr == 2:
                tx.item = v * 1.05
            elif rr == 3:
                tx.item = v + random.random()*v
            elif rr == 4:
                tx.item = v - random.random()*v
            else:
                tx.item = random.randint(-10,100)
            tx.item = round(tx.item,7)
        else:
            tx = makeitem(paradat)
            ##  cd also occasionally become a tree ?
    elif t.type == "stringy":
        assert len(t.item) == 2
        r = random.random()
        if r < 0.5:  ##  change constant
            fn,s = t.item  ##  unpack
            assert fn in paradat.strcols
            s = random.choice(paradat.strdict[fn])                
            tx.item = (fn,s)
            assert fn in paradat.cols2use
        elif r < 0.7:  ##  fresh stringy pair
            assert t.item[0] in paradat.strcols
            sset = set(paradat.strcols) & set(paradat.cols2use)
            if paradat.targcol in sset:  ## redundant ??
                sset.discard(paradat.targcol)
            assert len(sset) > 0
            fn = random.choice(list(sset))
            s = random.choice(paradat.strdict[fn])
            tx.item = (fn,s)  ##  feature,string held as pair
            assert fn in paradat.cols2use
        elif r < 0.9:  ##  fresh atom
            tx = makeitem(paradat)
        else:  ##  fresh tree
            tx = maketree(paradat,2+random.randint(1,2))
            ##  .type changed by subroutine
    elif t.type == "monadic":
        assert t.item in Monadics
        r = random.random()
        if r < 0.2:
            tx.item = random.choice(Monadics)
        elif r < 0.5:  ##  fresh argument
            tx.item = t.item
            tx.rh = zapnode(paradat,t.rh)
        elif r < 0.7:
            tx = makemono(paradat,2+random.randint(0,2))
        elif r < 0.9:
            tx = makeitem(paradat)
        else:
            tx = maketree(paradat,5)
    elif t.item in Dyadics:
        ##  r = random.random()
        r = random.randint(1,4)
        if r == 1:  ##  just pick another op
            if random.random() > 0.25:  ##  same type
                op = random.choice(Optable[t.type])
            else:  ##  free choice
                op = random.choice(Dyadics)
            tx.item = op
            tx.type = opclass(op)
        elif r == 2:  ##  make fresh expression
            funclist = [makebool,makecomp,makemath,maketree]
            typenum = ["bool","comp","math"].index(t.type)
            if random.random() < 0.5:
                func = funclist[typenum]  ##  bias towards same type
            else:  ##  free choice
                func = random.choice(funclist)
            tx = func(paradat,max(2,size(t) // 2))
        elif r == 3:  ##  clobber lh subtree
            tx.item = t.item
            tx.lh = zapnode(paradat,t.lh)
        else:  ##  clobber rh subtree
            tx.item = t.item
            tx.rh = zapnode(paradat,t.rh)
    else:
        tx.item = t.item  ##  no change
            
    return  tx


##  COPYZAP :
def copyzap (paradat,t,zn):
    "copying t with error at node zn."

    if not t:
        return  t  ##  null returns null
    tc = Sack()
    tc.lh = '' ; tc.rh = ''  ##  defaults
    if t.lh:
        tc.lh = copyzap(paradat,t.lh,zn)
    if t.rh:
        tc.rh = copyzap(paradat,t.rh,zn)
    if id(t) == id(zn):  ##  zn is node to mutate
        tc = zapnode(paradat,t)
    else:
        ##  copy core :
        tc.item = t.item
        tc.type = t.type

    return  tc
    ##  creates extra work for garbage collector ?


##  TREETOX :
def treetox (st):
    "function to take tree-string and return token list."

    betwixt = 0  ##  in between backtix
    buckseen = 0  ##  seen start of monadic op
    negseen = 0
    toks = []
    newt = ''
    s = st + ' '
    ##  at present Less must be surrounded by blanks when used as operator;
    ##  if followed by non-bracket, assumed 2b sign of a number.
    ##  prescan to detect violations probably unnecessary?
    for ch in s:
        if betwixt:  ##  inside tickquotes
            if ch == Backtick:  ##  ending
                toks.append(newt)  ##  keep closing char ?
                newt = ''
                betwixt = 0
            else:
                newt += ch
        elif buckseen:  ##  unary op started
            if ch in [Lbra,' ']:  ##  ending
                toks.append(newt)
                buckseen = 0
                if ch == Lbra:
                    toks.append(ch)
                newt = ''
            else:
                newt += ch
        elif negseen:  ##  cd be negative number
            if ch == ' ':  ##  dyadic (len(newt) must be 1)
                toks.append(newt)
                newt = ''
                negseen = 0
            elif ch == Rbra:  ##  negative number (len(newt) must be > 1)
                toks.append(newt)
                newt = ch
                negseen = 0
            else:
                newt += ch
        else:  ##  outside contexts
            if ch == Less:
                negseen = 1
                newt = ch
            elif ch == Buck:
                buckseen = 1
                newt = ch
            elif ch == Backtick:
                betwixt = 1
                newt = ch
            elif ch in [Lbra,Rbra]:
                if newt:
                    toks.append(newt)
                toks.append(ch)
                newt = ''
            elif ch in Dyadics:
                if newt:
                    toks.append(newt)
                toks.append(ch)
                newt = ''
            elif ch != ' ':
                newt += ch
            else:
                if newt:
                    toks.append(newt)
                    newt = ''
                
    ##  print(toks)
    lc = toks.count(Lbra) ; rc = toks.count(Rbra)
    ##  print(lc,rc)
    if lc != rc:
        print("Warning: unbalanced parentheses in",st)
    
    return  toks


##  OPERAND :
def operand (ts):
    "returns whether token-string ts is an operand."

    if len(ts) < 1:
        return  False
    n = numerize(ts)
    if numeric(n):
        return  True  ##  legitimate number
    if ts[0].isalpha():
        return  True  ##  variable name
    if ts[0] == Backtick:
        return  True  ##  string constant

    return  False


##  TOX2POLE :
def tox2pole (toklist):
    "converts infix token list to reverse Polish."
    ##  balanced brackets checked before entry.
    
    pole = []
    pile = []
    dyadics = Dyadics + [Contains]  ##  include 'in' pseudo-op
    for t in toklist:
        if len(t) < 1:
            continue  ##  nulls possible ??
        if operand(t):
            pole.append(t)
        elif t[0] == Buck:  ##  monadic
            pile.append(t)
        elif t in dyadics:  ##  dyadic
            ##  first check whether monadics waiting :
            while pile:
                if pile[-1][1:] in Monadics:  ##  higher priority
                    tt = pile.pop()
                    pole.append(tt)
                else:
                    break  ##  no more monadics
            pile.append(t)
        elif t == Lbra:
            pile.append(t)
        elif t == Rbra:
            while pile:
                tt = pile.pop()
                if tt == Lbra:
                    break
                pole.append(tt)
            ##  might still be some monadics in wait :
            while pile:
                if pile[-1][0] == Buck:
                ##  if pile[-1][1:] in Monadics:
                    tt = pile.pop()
                    pole.append(tt)
                else:
                    break  ##  no more monadics
        else:  ##  error ??
            print("??!!",t)
    ##  flush stack in case monadics left on it :
    while pile:
        tt = pile.pop()
        pole.append(tt)

    return  pole


##  POLETREE :
def poletree (paradat,rp):
    "converts reverse polish list rp to expression tree."

    pile = []
    for x in rp:
        if operand(x):  ##  make atomic node
            v = numerize(x)
            if numeric(v):  ##  number
                e = newnode('',v,'',"numeric")
            elif x and x[0] == Backtick:  ##  string constant
                e = newnode('',x[1:],'',"stringy")  ##  other type ??
            else:  ##  must be feature
                if x in paradat.taglist:
                    fn = paradat.taglist.index(x)
                    e = newnode('',fn,'',"feature")
                else:
                    raise RuntimeError("unknown feature "+x+"!!")
            pile.append(e)
        elif x == Contains:  ##  special case
            tl = pile.pop()
            hd = pile.pop() ; hd = hd.item
            e = newnode('',(hd,tl.item),'',"stringy")
            pile.append(e)
        elif x[1:] in Monadics:  ##  unary operator
            tl = pile.pop()
            e = newnode('',x[1:],tl,"monadic")
            pile.append(e)
        elif x in Dyadics:  ##  dyadic operator
            tl = pile.pop()
            hd = pile.pop()
            e = newnode(hd,x,tl,opclass(x))
            pile.append(e)
        else:  ##  oh dear !!
            print(rp)
            raise RuntimeError("can't interpret item "+x+"!!")

    if len(pile) != 1:  ##  problem
        print(rp)
        print(pile)
        raise RuntimeError("Trouble in pole-to-tree conversion !!")

    return  pile[-1]  ##  stacktop should be full tree


##  TRIGSET :
def trigset (trigflag):
    "includes or removes trig functions as indicated by trigflag."
    global  Monadics, Opchars

    if trigflag > 0:
        Monadics = [Isnt,Negative,"Cosi","Fabs","Root","Sine","Slog","Tanh"]
        Opchars = Dyadics + Monadics  ##  -> Opcodes
    else:
        Monadics = [Isnt,Negative,"Fabs","Root","Slog","Tanh"]
        Opchars = Dyadics + Monadics  ##  -> Opcodes

    return




Varnames = ["zero","unit","pair","four","luck","Minus99","nine",
            "Zero","sub1","half","text","name"]


##  test blocks :

if 0 > 0:  ##  testing
    print(Varnames)
    ##  pseudo-context :
    paradat = Sack()
    paradat.varnames = Varnames
    paradat.taglist = Varnames
    paradat.thisrow = [0,1,2,4,7,-99,9,0,-1,0.5,"Hello World!","Pythonic"]
    paradat.cols = len(paradat.varnames)
    assert len(paradat.thisrow) == paradat.cols
    paradat.allcols = list(range(paradat.cols))
    paradat.skipcols = set([0,8])  ##  forbidden
    paradat.goodcols = set(paradat.allcols) - paradat.skipcols
    paradat.strcols = [10,11]
    paradat.goodcols = paradat.goodcols - set(paradat.strcols)
    paradat.goodlist = sorted(paradat.goodcols)  ##  to save repeated re-listing
    ##  specimen string constants :
    paradat.strdict = {10: ["ello",'!',"zonk","lo","Hell","wo",'W'],
                       11: ["ic",'P','y',"zonq","xxzz",'@']}
    paradat.cols2use = sorted(paradat.goodlist + paradat.strcols)
    print(paradat.goodcols,paradat.goodlist,paradat.cols2use)

if 0 > 0:  ##  initial mate testing :
    m = maketree(paradat,7)
    p = maketree(paradat,7)
    print(NL,tree2str(m,paradat))
    print(NL,tree2str(p,paradat))
    x = mate(paradat,m,p)
    print("=>",tree2str(x,paradat))
    ##  nodelist = getnodes(x,paradat)
    ##  n = len(nodelist)
    ##  nz = nodelist[random.randint(0,n-1)]
    ##  print(n,tree2str(nz,paradat))
    ##  zt = zaptest(paradat,x,nz)
    xx = zonk(paradat,x)
    ##  xx = copytree(x)
    print(";;",tree2str(xx,paradat))
    print('=',treeval(xx,paradat,paradat.thisrow))
    xs = tree2str(x,paradat)
    print(xs)
    toks = treetox(xs)
    print(' '.join(toks))
    pole = tox2pole(toks)
    print(pole)
    pt = poletree(paradat,pole)
    ##  showobj(pt)
    ##  nodelist = getnodes(pt,paradat)
    ##  showitem(nodelist[0],paradat)
    ##  for j,n in enumerate(nodelist):
        ##  print(j) ; showobj(n)
    pout = tree2str(pt,paradat)
    print("::",pout)

if 0 > 0:
    ##  test tokenizer :
    ##  ti = "$Root (luck * -9)"
    ##  ti = "( $Slog((half ? `on`) - -.99))"
    ##  ti = "( (unit*7) - (( $Sine -99) *(half|-9)) )"
    ti = "( ($Fabs $Fabs 12.75 ) + $~ $~ (nine > nine))"
    toks = treetox(ti)
    print(ti)
    print(toks)
    pole = tox2pole(toks)
    to = poletree(paradat,pole)
    tout = tree2str(to,paradat)
    print(tout)

if 0 > 0:
    ##  try2 convert back :
    for j in range(64):
        t1 = maketree(paradat,6)
        if random.random() > 0.5:
            t1 = zonk(paradat,t1)
        v1 = treeval(t1,paradat,paradat.thisrow)
        ts = tree2str(t1,paradat)
        toks = treetox(ts)
        pole = tox2pole(toks)
        t2 = poletree(paradat,pole)
        v2 = treeval(t2,paradat,paradat.thisrow)
        print(v1,v2)
        if v1 != v2:
            print(ts)
            print("!!!!")
            break
        paradat.thisrow[4] = random.randint(-99,100)

if 0 > 0:
    if len(nodelist) > 2:
        nodelist[2] = maketree(paradat,5)
    for j,n in enumerate(nodelist):
        print(j,">>",tree2str(n,paradat))
    print(tree2str(x,paradat))
    if x.lh and x.rh:
        x.lh,x.rh = x.rh,x.lh
        print(tree2str(x,paradat))
    
if 0 > 0:
    nx = makeitem(paradat)
    sx = size(nx)
    print("atom :")
    print("size =",size(nx))
    print(tree2str(nx,paradat))
    print('=',treeval(nx,paradat,paradat.thisrow))
    showitem(nx,paradat)
    nx = makebool(paradat,7)
    print("bool :")
    print("size =",size(nx))
    print(tree2str(nx,paradat))
    print('=',treeval(nx,paradat,paradat.thisrow))
    showitem(nx,paradat)
    nx = makecomp(paradat,7)
    print("comp :")
    print("size =",size(nx))
    print(tree2str(nx,paradat))
    print('=',treeval(nx,paradat,paradat.thisrow))
    showitem(nx,paradat)
    nx = makemath(paradat,7)
    print("math :")
    print("size =",size(nx))
    print(tree2str(nx,paradat))
    print('=',treeval(nx,paradat,paradat.thisrow))
    showitem(nx,paradat)
    nx = maketree(paradat,7)
    print("tree :")
    print("size =",size(nx))
    print(tree2str(nx,paradat))
    print('=',treeval(nx,paradat,paradat.thisrow))
    showitem(nx,paradat)
    showobj(nx)
    nodelist = getnodes(nx,paradat)
    print(len(nodelist))
    print(nodelist)
    nodelist = getnodes(nx,paradat,False)
    print(len(nodelist))
    print(nodelist)
    print(paradat.thisrow)

if 0 > 0:
    ##  test copying of nested mutables :
    print()
    s1 = Sack(1)
    s1.hi = 99
    s1.lo = -99
    s1.svec = [1,2,3,4,5,['a','b'],7]
    ##  showobj(s1)
    s2 = copy.deepcopy(s1)
    s2.id = 2
    ##  showobj(s2)
    s1.svec[5] = "six"
    showobj(s1) ; showobj(s2)
    ##  deepcopy seems to handle mutables okay.
