##  Odds & ends in R, by R.S.Forsyth.
##  first version : 01/03/2011
##  last revision : 26/07/2016

Htab = '\t'
NL = '\n'
##  Gold = (1+sqrt(5)) * 0.5
Gold = 5.0 ^ 0.5 * 0.5 + 0.5
Tiny = 3e-33
Huge = 3e33



##  !! better 2use read.delim() :
datload = function (filename,dlim=Htab,heading=TRUE) {
	##  saves typing standard options for upload of delimited flat-file.

	df = read.table(filename,sep=dlim,header=heading,fileEncoding="UTF-8",
		quote='')

	return  (df)
	}

datdump = function (df,filename,dlim=Htab,heading=TRUE) {
	##  saves typing standard options when dumping dataframe as flat file.

	write.table(df,filename,sep=dlim,col.names=heading,quote=F,row.names=F)

	}

vecdist = function (vec1,vec2,p=2) {
	##  (euclidean-default) distance between 2 equal-length numeric vectors

	n1 = length(vec1) ; n2 = length(vec2)
	if (n1 != n2) {
		cat("Warning in vecdist() : unequal lengths !!\n")
		##  return (0)
		}
	n = min(n1,n2)  ##  should be the same
	difs = vec1[1:n] - vec2[1:n]
	difs = abs(difs) ^ p
	return  (sum(difs) ^ (1/p))
	}

rocarea = function (vec1,vec2) {
	##  area under ROC curve for 2 sets of scores.

	area = 0
	n1 = length(vec1) ; n2 = length(vec2)
	for (i in 1:n1) {
		sp = vec1[i]
		g = sum(sp > vec2)
		e = sum(sp == vec2)
		area = area + g + (0.5*e)
		}
	dnom = 1/(n1*n2)

	return (area * dnom)
	}

hedgesg = function (vec1,vec2) {
	##  Hedge's g, measure of effect size for 2 groups.

	n1 = length(vec1) ; n2 = length(vec2)
	if (((n1+n2) < 4) || ((n1*n2) < 1))
		return (0)  ##  no good
	diff = mean(vec1) - mean(vec2)
	if ((n1+n2) < 7)  ##  puny sample
		poolvar = max(var(vec1),var(vec2))
	else  ##  normal case
		poolvar = ((n1-1)*var(vec1)+(n2-1)*var(vec2)) / (n1+n2-2)

	return (diff / sqrt(poolvar))
	}

ssqratio = function (xvec,cvec) {
	##  crude R-squared as in 1-way anovar.

	nx = length(xvec)
	if ((nx<3) || (nx != length(cvec)))
		return (0)  ##  can't be done
	catcodes = intersect(cvec,cvec)  ##  unique values (in order)
	cats = length(catcodes)
	xbar = mean(xvec)  ##  grand mean
	ssqt = sum((xvec-xbar)^2)
	cseq = 1:cats  ##  catcodes mightn't be consecutive
	ssqc = rep(0,cats)  ##  group sums of squared devs.
	##  cat(catcodes,NL,cseq,NL)
	for (c in cseq) {
		cc = catcodes[c]
		v = xvec[cc == cvec]
		ssqc[c] = sum((v-mean(v))^2)
		}
	ssqw = sum(ssqc)
	cat(ssqt,NL)
	##  print(ssqc)
	return (1 - ssqw / max(1e-16,ssqt))
	}

mindist = function (dmat,t5=4,tagsize=2) {  ##  near-neighbour stats on a full distance matrix

	n = dim(dmat)[1]
	top5 = min(t5,n)
	nvec = colnames(dmat)  ##  must have column names
	nearvec = character(0)
	mm = 0  ##  mismatches
	for (c in 1:n)  {  ##  for each column
		cvec = dmat[,c]
		cord = order(cvec)  ##  smallest at front
		min4 = cvec[cord[1:top5]]
		n4 = substr(nvec[cord[1:top5]],1,8)
		cat(c,min4,n4,NL)
		nearvec[c] = substr(n4[2],1,tagsize)  ##  skip self
		if (substr(n4[1],1,tagsize) != substr(n4[2],1,tagsize))  mm = mm + 1
		}
	print(mm)
	return  (nearvec)
	}


binadd1 = function (bvec) {  ##  adds 1 to quasi-binary vector

	nb = length(bvec)
	if (all(bvec>0)) {  ##  quasi-overflow
		newb = rep(0,nb)
		newb[nb+1] = 1
		return (newb)
		}
	##  low values at left (lower indices) :
	p = 1
	bvec[p] = bvec[p] + 1
	while (bvec[p] > 1) {  ##  have to carry
		bvec[p] = 0
		p = p + 1
		bvec[p] = bvec[p] + 1
		}
	## if length(bvec) > nb, signal to stop adding.
	return (bvec)
	}


zedness = function (points1,points2,catlist1,catlist2=NULL,crosspos=NULL) {
	##  catlists give group membership for points in respective matrices;
	##  crosspos[i] gives position of twin in points2 for points1[i]:
	##  give 1:rows when self-testing with both mats the same.

	cols = dim(points1)[2]
	if (cols != dim(points2)[2])  stop("Matrices don't match !!")
	n1 = dim(points1)[1] ; n2 = dim(points2)[1]
	if (n1 < 1 || n2 < 1)  stop("Matrices too tiny !!")
	diffvec = samevec = twinvec = c()
	alldiffs = numeric(max(n1,n2))
	if (length(catlist2)==0)  cat2 = catlist1
	##  assumes same groupings if none given.
	cat(catlist2,NL)
	if (length(crosspos)==0)  crosspos = 1:min(n1,n2)
	cat(crosspos,NL)
	nd = 0
	for (i in 1:n1) {
		j = i
		while (j <= n2) {
			nd = nd + 1
			d = vecdist(points1[i,],points2[j,])
			if (i == crosspos[j])
				twinvec = c(twinvec,d)
			else if (catlist1[i]==catlist2[j])
				samevec = c(samevec,d)
			else
				diffvec = c(diffvec,d)
			alldiffs[nd] = d  ##  to save time
			j = j + 1
			}
		}
	m1 = mean(diffvec) ; m2 = mean(samevec) ; m3 = mean(twinvec)
	cat(m1,m2,m3,NL)
	znum = (m1 - m2) + (m2 - m3)
	znum = hedgesg(diffvec,samevec) + hedgesg(samevec,twinvec)
	return (znum)
	##  values should be ascending.
	}

groupdis = function (points1,catlist1,points2=NULL,catlist2=NULL,crosspos=NULL) {
	##  extracts distances in 2/3 groups from matrices of coordinates:
	##  crosspos gives position of twin in points2 for points1[i], 0 for none.
	##  defaults to 1:rows in self-test mode.

	selftest = F
	cols = dim(points1)[2]
	if (length(points2) < 1)  {
		points2 = points1 ; selftest = T
		}
	if (cols != dim(points2)[2])  stop("Matrices don't match !!")
	n1 = dim(points1)[1] ; n2 = dim(points2)[1]
	if (n1 < 1 || n2 < 1)  stop("Matrices too tiny !!")
	diffvec = samevec = twinvec = c()
	if (length(catlist2)==0)
		cat2 = catlist1  ##  both point sets have same groupings
	else  cat2 = catlist2  ##  separate category lists
	cat(cat2,NL)
	if (length(crosspos)==0)  crosspos = 1:min(n1,n2)
	cat(crosspos,NL)
	twinvec = numeric(n1) ; tc = 0
	samevec = numeric(max(n1,n2)) ; sc = 0
	diffvec = numeric(n1*n2 / 2) ; dc = 0
	nd = 0
	for (i in 1:n1) {
		j = i
		while (j <= n2) {
			if (selftest && i==j)  next  ##  twins don't count when in self-test mode
			d = vecdist(points1[i,],points2[j,])
			if (i == crosspos[j]) {  ##  closest pair
				tc = tc + 1
				twinvec[tc] = d
				}
			else if (catlist1[i]==cat2[j]) {  ##  pair in same group
				sc = sc + 1
				samevec[sc] = d
				}
			else  {  ##  pair in different groups
				dc = dc + 1
				diffvec[dc] = d
				}
			j = j + 1
			}
		}
	print(tc)
	if (tc < 1 || selftest)
		return (list(samevec[1:sc],diffvec[1:dc]))
	else
		return (list(twinvec[1:tc],samevec[1:sc],diffvec[1:dc]))
	}  ##  ready to evaluate with moes().


moes = function (veclist) {  ## Mean Ordinal Effect Size:
	##  expects a list of numeric vectors.

	nv = length(veclist)
	if (nv < 2)  return (0)  ##  needs more than 1 group
	htot = 0
	for (g in 2:nv) {
		hg = hedgesg(veclist[[g]],veclist[[g-1]])
		##  ascending order presumed.
		htot = htot + hg
		}

	return (htot / (nv-1))
	}

safecorr = function (xvec,yvec,epsilon=3e-33) {
	##  computes r with special cases if either variance is zero.

	vx = var(xvec) ; vy = var(yvec)
	if ((vx < epsilon) & (vy < epsilon))
		return (1)  ##  neither varies
	if ((vx < epsilon) | (vy < epsilon))
		return (0)  ##  1 varies but not both
	##  normal case, both vectors have some variance :
	return (cov(xvec,yvec) / sqrt(vx*vy))
	}

shannent = function (pvec,entbase=2) {
	##  computes mean entropy of a vector of proportions

	tp = sum(pvec)
	if (abs(tp-1) > Tiny)  stop("p-vector must sum to 1 !")
	np = length(pvec)
	tent = 0.0
	for (pv in pvec) {
		##  cat(pv,tent)
		if (pv > Tiny)  tent = tent + pv * log(pv,entbase)
		##  cat(' ',tent,NL)
		}
	return  (-tent)
	}

probmod = function (pvec) {
	##  modifies probability vector by adding a "none-of-the-above" item
	##  uses relative entropy to guess prob of dustbin class, then appends it & rescales
	ve = shannent(pvec,2)
	np = length(pvec)
	if (np < 2)  ##  silly data
		xp = Gold
	else {
		xp = ve / (-log(1/np,2))  ##  relative entropy
		xp = xp / Gold
		}
	tote = sum(pvec) + xp
	return (c(pvec,xp)/tote)
	}  ##  probabilities = proportions - VAT  ;-)

hausmed = function (aa,bb,dfun=abs) {
	##  medianized Hausdorff distance between point sets.

	na = length(aa)
	md = numeric(na)
	for (j in 1:na) {
		a = aa[j]
		dv = abs(a-bb)  ##  a vector
		md[j] = median(dv)
		}
	return (median(md))
	}  ##  superseded by hausdist().

cgrfunc1 = function (st,frac=0.5) {
	##  chaos-game representation of a string; most basic version.
	##  returns a dataframe.

	if (frac < Tiny || frac > 1.0)  stop("frac must be (0,1).")
	punk = c('.',',','!','?','-',"'",':',';')
	vv = c('a','e','i','o','u','y')
	cc = setdiff(letters,vv)
	baseline=list(c(0,0),c(0,1),c(1,0),c(1,1))
	p = c(frac,1-frac)
	ns = nchar(st)
	horz = p[1] ; vert = p[2]
	for (j in 1:ns) {
		ch = substr(st,j,j)
		ct = 4
		if (ch %in% punk)  ct = 1
		if (ch %in% cc)  ct = 2
		if (ch %in% vv)  ct = 3
		goal = baseline[[ct]]
		p[1] = goal[1]*frac + p[1]*(1-frac)
		p[2] = goal[2]*frac + p[2]*(1-frac)
		horz = c(horz,p[1])
		vert = c(vert,p[2])
		}

	df = data.frame(x=horz,y=vert)
	return (df)
	}


cgrfunc2 = function (st,frac=0.5) {
	##  chaos-game representation of a string; most basic version.
	##  returns a dataframe.

	if (frac < Tiny || frac > 1.0)  stop("frac must be (0,1).")
	punk = c('.',',','!','?','-',"'",':',';')
	vv = c('a','e','i','o','u','y')
	cc = setdiff(letters,vv)
	vu = toupper(vv) ; cu = tolower(cc)
	digs = c('0','1','2','3','4','5','6','7','8','9')
	baseline=list(c(0,0),c(0,0.5),c(0,1),c(0.5,0),c(0.5,1),c(1,0),c(1,0.5),c(1,1))
	p = c(frac,1-frac)
	ns = nchar(st)
	horz = p[1] ; vert = p[2]
	for (j in 1:ns) {
		ch = substr(st,j,j)
		ct = 8
		if (ch == ' ')  ct = 7
		if (ch %in% punk)  ct = 1
		if (ch %in% cc)  ct = 2
		if (ch %in% vv)  ct = 3
		if (ch %in% cu)  ct = 4
		if (ch %in% vu)  ct = 5
		if (ch %in% digs)  ct = 6
		goal = baseline[[ct]]
		p[1] = goal[1]*frac + p[1]*(1-frac)
		p[2] = goal[2]*frac + p[2]*(1-frac)
		horz = c(horz,p[1])
		vert = c(vert,p[2])
		}

	df = data.frame(x=horz,y=vert)
	return (df)
	}


hausdis0 = function (dat1,dat2,pval=2,qpos=1) {
	##  Hausdorff distance between 2 point-sets, held as data frames.
	##  pval selects distance mode (L); qpos (0..1) selects quantile mode.
	##  both passed to hausfun().

	if (qpos < 0 || qpos > 1) {
		cat("Warning: qpos must be 0..1; limit imposed!")
		qpos = max(0,qpos)
		qpos = min(1,qpos)
		}
	h1 = hausfun(dat1,dat2,pval,qpos)
	h2 = hausfun(dat2,dat1,pval,qpos)
	##  cat(h1,h2,NL)
	return (max(h1,h2)*qpos + min(h1,h2)*(1-qpos))  ##  weighted as in quantilization.
	}


hausdist = function (dat1,dat2,pval=2,qpos=1) {
	##  revised Hausdorff distance between 2 point-sets, held as data frames.
	##  pval selects distance mode (L); qpos (0..1) selects quantile mode.
	##  both passed to hausfun().

	if (qpos < 0 || qpos > 1) {
		cat("Warning: qpos must be 0..1; limit imposed!")
		qpos = max(0,qpos)
		qpos = min(1,qpos)
		}
	n1 = dim(dat1)[1] ; n2 = dim(dat2)[1]
	if (n1 < 2 || n2 < 2) 
		stop("datframes don't have enough rows !")
	##  pre-compute item-based distance matrix :
	dmat = matrix(nrow=n1,ncol=n2)
	for (i in 1:n1) {
		a = t(dat1[i,])
		for (j in 1:n2) {
			dmat[i,j] = vecdist(a,t(dat2[j,]),pval)
			}
		}
	h1 = hausfun(dmat,qpos)
	h2 = hausfun(t(dmat),qpos)
	cat(h1,h2,NL)
	return (max(h1,h2)*qpos + min(h1,h2)*(1-qpos))  ##  weighted as in quantilization.
	}
	
	
hausfun = function (dmat,qpos=1) {
	##  directed Hausdorff distance function.
	##  uses vecdist() on each row.

	r1 = dim(dmat)[1] ; r2 = dim(dmat)[2]
	dmax = -Inf
	mins = numeric(r1)
	for (i in 1:r1) {
		dvec = numeric(r2)
		for (j in 1:r2) {
			dvec[j] = dmat[i,j]
			}  ##  also quantilized.
		dmin = quantile(dvec,1-qpos,na.rm=T)
		mins[i] = dmin  ##  minimum in default case; generally lowish quantile
		##  cat(i,pmin,a,bmin,dmin,NL)
		##  print(a) ; print(bmin)
		}
	h = quantile(mins,qpos,na.rm=T)  ##  maximum in default case; generally highish quantile
	return (h)
	}
	
	
hausfun0 = function (dat1,dat2,pval=2,qpos=1) {
	##  directed Hausdorff distance function.
	##  uses vecdist() on each row.

	r1 = dim(dat1)[1] ; r2 = dim(dat2)[1]
	dmax = -Inf
	mins = numeric(r1)
	for (i in 1:r1) {
		dmin = Inf
		a = t(dat1[i,])  ##  coerce to column vector
		dvec = numeric(r2)
		for (j in 1:r2) {
			b = t(dat2[j,])  ##  coerce to vector
			d = vecdist(a,b,pval)
			dvec[j] = d
			}  ##  also quantilized.
		dmin = quantile(dvec,1-qpos,na.rm=T)
		mins[i] = dmin  ##  minimum in default case; generally lowish quantile
		##  cat(i,pmin,a,bmin,dmin,NL)
		##  print(a) ; print(bmin)
		}
	h = quantile(mins,qpos,na.rm=T)  ##  maximum in default case; generally highish quantile
	return (h)
	}

haustest = function (dat1,dat2,pval=2,qpos=1) {
	##  test of hausdist minimization.
	
	n1 = dim(dat1)[1] ; n2 = dim(dat2)[1]
	num1 = numeric(0) ; num2 = numeric(0)
	hmin = Inf ; hvec = numeric(0)
	for (j in 1:999) {
		##  h1 = round(n1/2) ; sam1 = sample(1:n1,h1)
		##  h2 = round(n2/2) ; sam2 = sample(1:n2,h2)
		##  approx half at random :
		sam1 = runif(n1) ; sam2 = runif(n2)
		sam1 = sam1 > 0.5 ; sam2 = sam2 > 0.5
		h1 = sum(sam1) ; h2 = sum(sam2)
		hd = hausdist(dat1[sam1,],dat2[sam2,],pval,qpos)
		num1[j] = h1 ; num2[j] = h2
		hvec[j] = hd
		##  print(sam1) ; print(sam2)
		if (hd < hmin) {
			hmin = hd
			hdat = rbind(dat1[sam1,],dat2[sam2,])
			hdat$type = c(rep(1,h1),rep(2,h2))
			}
		}
	hval = data.frame(num1,num2,num1+num2,hvec)
	return (list(hval,hdat))
	}

	
hausopt = function (dat1,dat2,pval=2,qpos=1) {
	##  test of hausdist optimization.
	
	n1 = dim(dat1)[1] ; n2 = dim(dat2)[1]
	sam1 = rep(T,n1) ; sam2 = rep(T,n2)
	nn = n1 + n2
	postdist = numeric(nn)
	removed = numeric(nn)
	stayers = nn
	dmat = prehaus(dat1,dat2,pval,qpos)
	while (stayers > max(n1,n2)) {
		##  predist[stayers] = hausdist(dat1[sam1,],dat2[sam2,],pval,qpos)
		hvec = numeric(0) ; hpos = numeric(0)
		for (j in 1:nn) {
			##  try removing item j :
			jj = j
			if (j > n1) {  ##  dat2 involved
				jj = j - n1
				if (sam2[jj] == F)  next  ##  already masked out
				}
			else {
				if (sam1[jj] == F)  next  ##  already masked out
				}
			if (j <= n1)  {
				sam1[jj] = F
				hd = posthaus(dmat[sam1,sam2],qpos)
				}
			else  {
				sam2[jj] = F
				hd = posthaus(dmat[sam1,sam2],qpos)
				}
			##  hd = posthaus(dmat,qpos)
			hvec = c(hvec,hd)
			hpos = c(hpos,j)
			##  restore mask vectors :
			if (j <= n1)  sam1[jj] = T
			else  sam2[jj] = T
			}
		wm = which.min(hvec)
		mp = hpos[wm]
		##  confirm removal of 1 item :
		if (mp <= n1)  sam1[mp] = F
		else  sam2[mp-n1] = F
		postdist[stayers] = hvec[wm]
		removed[stayers] = mp
		stayers = stayers - 1
		cat(stayers,mp,hvec[wm],NL)
		}
		
	return (list(sam1,sam2,removed,postdist))
	}

prehaus = function (dat1,dat2,pval=2,qpos=1) {
	##  pre-computes inter-item distance matrix from 2 data frames.
	
	n1 = dim(dat1)[1] ; n2 = dim(dat2)[1]
	if (n1 < 2 || n2 < 2) 
		stop("datframes don't have enough rows !")
	##  pre-compute item-based distance matrix :
	dmat = matrix(nrow=n1,ncol=n2)
	for (i in 1:n1) {
		a = t(dat1[i,])
		for (j in 1:n2) {
			dmat[i,j] = vecdist(a,t(dat2[j,]),pval)
			}
		}
	return (dmat)
	}

posthaus = function (dmat,qpos=1) {
	##  computes hausdorff distance from matrix of inter-item distances.
	
	if (qpos < 0 || qpos > 1) {
		cat("Warning: qpos must be 0..1; limit imposed!")
		qpos = max(0,qpos)
		qpos = min(1,qpos)
		}
	h1 = hausfun(dmat,qpos)
	h2 = hausfun(t(dmat),qpos)
	##  cat(h1,h2,NL)
	return (max(h1,h2)*qpos + min(h1,h2)*(1-qpos))  ##  weighted as in quantilization.
	}

	
koncord = function (kmat) {
	##  Kendall's coefficient of concordance (uncorrected for ties).
	
	rc = dim(kmat)
	rows = rc[1] ; cols = rc[2]
	##  rows ~ judges ; cols ~ items.
	if (rows < 2 || cols < 2)  return  (0.0)
	
	colsums = numeric(cols)
	for (c in 1:cols)
		colsums[c] = sum(kmat[,c])
	colmean = mean(colsums)
	cssq = sum((colsums-colmean)^2)
	
	##  print(colsums)
	##  cat(colmean,cssq,NL)
	maxcssq = (rows^2 * (cols^3 - cols)) / 12
	
	return (cssq/maxcssq)
	}
	

konkold = function (kmat) {
	##  (older) Kendall's coefficient of concordance (adapted for ratings).
	
	rc = dim(kmat)
	rows = rc[1] ; cols = rc[2]
	##  rows ~ judges ; cols ~ items.
	if (rows < 2 || cols < 2)  return  (0.0)
	
	colsums = numeric(cols)
	kram = matrix(rank(kmat),nrow=rows)
	for (c in 1:cols)
		##  colsums[c] = sum(kmat[,c])
		colsums[c] = sum(kram[,c])
		##  colsums[c] = mean(kram)*rows
	colmean = mean(colsums)
	cssq = sum((colsums-colmean)^2)
	ranx = numeric(0)
	for (r in 1:rows) {
		rx = rank(kmat[r,])
		ranx = c(ranx,rx)
		}
	ranx = matrix(ranx,nrow=rows)
	sortvec = sort(as.vector(rank(kmat)))
	xmat = matrix(0,nrow=rows,ncol=cols)
	##  distribute rowwise :
	r = 1 ; c = 1
	incr = 1
	for (x in sortvec) {
		xmat[r,c] = mean(kram)  ##  x
		c = c + incr
		if (c > cols) {
			incr = -incr
			c = c + incr
			r = r + 1
			}
		else if (c < 1) {
			incr = - incr
			c = c + incr
			r = r + 1
			}  ##  boustrophedon
		}
	print(xmat)
	for (c in 1:cols)
		colsums[c] = sum(xmat[,c])
	colmean = mean(colsums)
	xssq = sum((colsums-colmean)^2)  ##  minimum
	##  cat(colmean,xssq,NL)
	##  distribute columnwise :
	r = 1 ; c = 1
	incr = 1
	for (x in sortvec) {
		xmat[r,c] = x
		r = r + incr
		if (r > rows) {
			incr = -incr
			r = r + incr
			c = c + 1
			}
		else if (r < 1) {
			incr = - incr
			r = r + incr
			c = c + 1
			}  ##  boustrophedon
		}
	print(xmat)
	for (c in 1:cols)
		colsums[c] = sum(xmat[,c])
	colmean = mean(colsums)
	yssq = sum((colsums-colmean)^2)  ##  maximum
	cat(colmean,xssq,yssq,NL)
	
	kw = (cssq - xssq) / (yssq - xssq)
	return (kw)
	}
	

konkord = function (kmat,byranks=F) {
	##  Kendall's coefficient of concordance (adapted for ratings).
	
	rc = dim(kmat)
	rows = rc[1] ; cols = rc[2]
	##  rows ~ judges ; cols ~ items.
	if (rows < 2 || cols < 2)  return  (0.0)
	
	colsums = numeric(cols)
	if (byranks) {  ##  ranked values
		##  rmat = rank(kmat)
		rmat = matrix(0,nrow=rows,ncol=cols)
		for (r in 1:rows)
			rmat[r,] = rank(kmat[r,])
		}
	else
		rmat = kmat  ##  raw data
	av = mean(rmat)
	for (c in 1:cols)
		colsums[c] = sum(rmat[,c])
	colmean = mean(colsums)
	cssq = sum((colsums-colmean)^2)
	ranx = numeric(0)
	sortvec = sort(rmat)
	xmat = matrix(0,nrow=rows,ncol=cols)
	##  distribute rowwise :
	r = 1 ; c = 1
	incr = 1
	##  cat(length(sortvec),dim(rmat),NL)
	for (x in sortvec) {
		##  cat(r,c,x,NL)
		xmat[r,c] = av  ##  mean(kram)  ##  x
		c = c + incr
		if (c > cols) {
			incr = -incr
			c = c + incr
			r = r + 1
			}
		else if (c < 1) {
			incr = -incr
			c = c + incr
			r = r + 1
			}  ##  boustrophedon
		}
	##  print(xmat)
	for (c in 1:cols)
		colsums[c] = sum(xmat[,c])
	colmean = mean(colsums)
	xssq = sum((colsums-colmean)^2)  ##  minimum
	##  always zero !
	##  distribute columnwise :
	r = 1 ; c = 1
	incr = 1
	for (x in sortvec) {
		xmat[r,c] = x
		r = r + incr
		if (r > rows) {
			incr = -incr
			r = r + incr
			c = c + 1
			}
		else if (r < 1) {
			incr = - incr
			r = r + incr
			c = c + 1
			}  ##  boustrophedon
		}
	##  print(xmat)
	for (c in 1:cols)
		colsums[c] = sum(xmat[,c])
	colmean = mean(colsums)
	yssq = sum((colsums-colmean)^2)  ##  maximum
	##  cat(colmean,cssq,xssq,yssq,NL)
	
	kw = (cssq - xssq) / (yssq - xssq)
	return (kw)
	}
	

propconc = function (vec1,vec2,skipties=T) {  ##  brute-force version
	##  proportion of concordant pairs (related to Kendall's tau & G&K lambda) :
	
	n1 = length(vec1) ; n2 = length(vec2)
	if (n1 < 1) {
		cat("Problem in propconc() : too few data points!",NL)
		return (0)
		}
	if (n1 != n2) {
		cat("Problem in propconc() : vector lengths differ :",n1,n2,"!",NL)
		return (0)
		}
	c = 0 ; d = 0
	ties = 0
	for (i in 1:n1) {
		x = vec1[i]
		for (j in i:n2) {
			if (i >= j)  next  ##  skip self-comparison
			d1 = x - vec1[j]
			d2 = vec2[i] - vec2[j]
			if (d1 == 0 && d2 == 0) c = c + 1
			else if (d1*d2 > 0)  c = c + 1
			else  d = d + 1
			}
		}
	cat(c,d,ties,c+d+ties,NL)
	if (skipties)  pc = c / max(0.5,c+d)
	else  pc = (c + 0.5*ties) / max(0.5,c+d+0.5*ties)
	
	return (pc)
	}

	
gastave <- function(x,...){
	##  Gastwirth's robust location estimator, from RonPearson :
	ordstats = quantile(x, probs=c(1/3,1/2,2/3),...)
	w = c(0.3,0.4,0.3)
	
	return (sum(w*ordstats))
    }
	
	
kagdist = function (xvec,yvec,logbase=0) {
	##  distance function for skewed counts (from Kaggle contest 2013) :
	
	if (logbase <= 0)
		e = mean((log(xvec+1) - log(yvec+1)) ^ 2)
	else
		e = mean((log(xvec+1,logbase) - log(yvec+1,logbase)) ^ 2)
		
	return  (sqrt(e))
	}


ssqdevs = function (xvec) {
	##  simple sum of squared difs from mean :

	n = length(xvec)
	if (n < 1)  return (0.0)
	m = mean(xvec)

	return (sum((m-xvec)^2))
	}
	##  divide by n to get uncorrected variance.

	
leafnum = 0  ##  still needs a global variable. :-(
clabhues = 0  ##  ditto : cluster label hues, 1 per leaf; global 2b preset. :-(

leafhues = function (node) {
	##  adds colour to a dendrogram leaf node :
	if (is.leaf(node)) {
		a = attributes(node)
		leafnum <<- leafnum + 1
		leafhue = clabhues[leafnum]
		##  cat(leafnum,leafhue,NL)
		attr(node,"nodePar") = c(a$nodePar,lab.col=leafhue,pch=NA)
		##  print(attributes(node))
		}
	node
	}
		

clushues = function (distmat,labtypes,labnames=character(0),horz=T,m="ward") {
	##  plots dendrogram with coloured leaves :
	##  labtypes (classes) determine hues; labnames give names (=labtypes by default).
	dm = as.dist(distmat)  ## ensure consistent format
	rows = attr(dm,"Size")
	if (length(labnames) == rows)  ##  treat labnames as leaf labels
		attr(dm,"Labels") = substr(labnames,1,16)
	else  {  ##  use type names as leaf labels (must be as many as leaves)
		attr(dm,"Labels") = labtypes
		if (length(labtypes) != rows)
			warning("wrong number of dendro-label type-codes!")
		}
	clus = hclust(dm,meth=m)
	##  number of hues should equal rows/cols in distmat.
	clustree = as.dendrogram(clus,hang=0.125)  ##  0?
	leafnum <<- 0
	##  clabhues <<- hues[clus$order]
	##  clabhues <<- hues
	##  establish leaf-label hues :
	cats = unique(labtypes)
	nh = length(cats)
	if (nh <= 13)
		hues = rainbow(nh)  ##  no good with more than c.13 cats
	else
		hues = rep(rainbow(13),length.out=nh)
	labs = length(labtypes)
	clabhues <<- rep(1,labs)
	for (h in 1:labs) {
		j = clus$order[h]
		p = match(labtypes[j],cats)
		clabhues[h] <<- hues[p]
		##  clustree$label = labtypes[j]
		}
	##  preset node numbering global :
	leafnum <<- 0
	huedtree = dendrapply(clustree,leafhues)
	plot(huedtree,horiz=horz)
	return  (huedtree)  ##  later cd be plotted vertically if required.
	}

prie = function (prop,ep) {
	##  proportional reduction in error.
	
	return (prop-ep) / (1-ep)
	}


tetracor = function (vec1,vec2,slug=0.5) {
	##  tetrachoric correlation between 2 vectors,
	##  which must be equal length & binary !!
	
	n = length(vec1)
	if (n != length(vec2))  stop("vector lengths differ in tetracor !!")
	a = sum(vec1==1 & vec2==1) + slug
	b = sum(vec1==1 & vec2==0) + slug
	c = sum(vec1==0 & vec2==1) + slug
	d = sum(vec1==0 & vec2==0) + slug
	##  nn = a+b+c+d
	##  slug > 0 avoids division by zero; slug==0 gives true coefficient.
	tv = (sqrt(a*d)-sqrt(b*c)) / (sqrt(a*d)+sqrt(b*c))
	
	return  (sin(pi*0.5 * tv))  ##  subtract from 1 if distance wanted.
	}
	
giniproc <- function(pvect){
  #
  #  from companion website of R.Pearson's Exploring Data book.
  #  Procedure to compute the normalized Gini heterogeneity measure
  #
  #  Note: the input variable pvect is assumed to be a normalized 
  #        probability vector that sums to 1 and is of length M.
  #        Missing values are not allowed, but components of pvector
  #        are allowed to be zero.
  #
  M <- length(pvect)
  if (M == 1){
    IGini <- 0
  }
  else{
    prefactor <- (M+1)/(M-1)
    wts <- (2*seq(1,M,1))/(M+1) - 1
    psort <- sort(pvect)
    rawsum <- sum(wts*psort)
    IGini <- prefactor * rawsum
    #
  }
  round(IGini,digits=7)
  #
}

gini.mean.diff = function (x) {
	##  Gini mean difference :
	
	n = length(x)
	if (n < 2)  return (0)
	y = sort(x)
	na = seq((1-n),(n-1),by=2)
	2 * sum(na*y) / n / (n-1)
	}  ##  multiply by 0.88622692545 to approximate sd scale of gaussian
##  as from Robust Stat Methods with R by Jureckova & Picek.

deltawt = function (dvec,fixmode=0) {
	##  routine to turn vector of deltas into weight vector (of quasi-probabilities) :
	nd = length(dvec) ; base = min(dvec) / 2
	if (nd < 2) stop("singleton vector given to deltawt")
	dvec = dvec + (dvec<1e-7)*1e-7  ##  ensure no zeroes
	if (fixmode==1)
		wvec = (dvec-base)
	else if (fixmode==2)
		wvec = (dvec+rank(dvec)^0.5)
	else if (fixmode==3) {
		medd = median(dvec) - dvec;
		wvec = dvec - medd
		}
	else  ##  just use deltas
		wvec = dvec
	wvec = 1.0 / wvec ^ 2
	return (wvec/sum(wvec))
}


##  text2tox :
text2tox = function (docfile,casefold=TRUE) {
	##  tokenizes 1 text file in docfile, resulting from docfiles=choose.files()
	
	docscan=scan(file=docfile,what=character(0),sep='\n')
	##  crude version, just uses predefined pattern 
	toks=strsplit(docscan, "\\W")
	head(toks) ; tail(toks)
	tokvect=unlist(toks)
	tokvect = tokvect[tokvect!='']  ##  remove null tokens (punctuation et cetera)
	if (casefold)
		tokvect=tolower(tokvect)
	return (tokvect)
	}

	
cortoks = function (tokfrex,toks) {
	##  correlation of a freq table with a vector of tokens :
	
	tf = table(toks)  ##  local freq table
	##  print(head(tf,12))
	##  print(head(names(tf),16))
	nf = length(tokfrex)
	##  cat(nf,tf["and"],NL)
	v = rep(0,nf)
	for (j in 1:nf) {
		n = names(tokfrex)[j]
		fx = tf[n]
		##  cat(j,n,fx,NL)
		if (is.na(fx))  v[j] = 0
		else  v[j] = fx
		}
	return(cor(tokfrex,v,meth="sp"))
	}
	
pickvect = function (datf,holdcat,holdouts) {
	##  returns T/F vector to pick entire class plus random test examples from datf.
	rows = nrow(datf) ; cols = ncol(datf)
	hout = min(holdouts,rows-1)
	hout = max(1,hout)
	selector = holdcat  ##  tf expr already input, selecting holdout class
	priority = order(runif(rows))
	##  print(priority)
	pict = sum(selector)
	for (p in priority) {
		if (pict >= hout)  return (selector)
		if (! selector[p]) {
			selector[p] = TRUE
			pict = pict + 1
			##  cat(p,pict,NL)
			}
		}
	return (selector)
	}
